# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Post-processing operations on detected boxes."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import numpy as np
from six.moves import range
from six.moves import zip
import tensorflow.compat.v1 as tf

from object_detection.core import box_list
from object_detection.core import box_list_ops
from object_detection.core import keypoint_ops
from object_detection.core import standard_fields as fields
from object_detection.utils import shape_utils

_NMS_TILE_SIZE = 512


def batch_iou(boxes1, boxes2):
  """Calculates the overlap between proposal and ground truth boxes.

  Some `boxes2` may have been padded.  The returned `iou` tensor for these
  boxes will be -1.

  Args:
    boxes1: a tensor with a shape of [batch_size, N, 4]. N is the number of
      proposals before groundtruth assignment. The last dimension is the pixel
      coordinates in [ymin, xmin, ymax, xmax] form.
    boxes2: a tensor with a shape of [batch_size, MAX_NUM_INSTANCES, 4]. This
      tensor might have paddings with a negative value.

  Returns:
    iou: a tensor with as a shape of [batch_size, N, MAX_NUM_INSTANCES].
  """
  with tf.name_scope('BatchIOU'):
    y1_min, x1_min, y1_max, x1_max = tf.split(
        value=boxes1, num_or_size_splits=4, axis=2)
    y2_min, x2_min, y2_max, x2_max = tf.split(
        value=boxes2, num_or_size_splits=4, axis=2)

    # Calculates the intersection area.
    intersection_xmin = tf.maximum(x1_min, tf.transpose(x2_min, [0, 2, 1]))
    intersection_xmax = tf.minimum(x1_max, tf.transpose(x2_max, [0, 2, 1]))
    intersection_ymin = tf.maximum(y1_min, tf.transpose(y2_min, [0, 2, 1]))
    intersection_ymax = tf.minimum(y1_max, tf.transpose(y2_max, [0, 2, 1]))
    intersection_area = tf.maximum(
        (intersection_xmax - intersection_xmin), 0) * tf.maximum(
            (intersection_ymax - intersection_ymin), 0)

    # Calculates the union area.
    area1 = (y1_max - y1_min) * (x1_max - x1_min)
    area2 = (y2_max - y2_min) * (x2_max - x2_min)
    # Adds a small epsilon to avoid divide-by-zero.
    union_area = area1 + tf.transpose(area2,
                                      [0, 2, 1]) - intersection_area + 1e-8

    # Calculates IoU.
    iou = intersection_area / union_area

    # Fills -1 for padded ground truth boxes.
    padding_mask = tf.logical_and(
        tf.less(intersection_xmax, 0), tf.less(intersection_ymax, 0))
    iou = tf.where(padding_mask, -tf.ones_like(iou), iou)

    return iou


def _self_suppression(iou, iou_threshold, loop_condition, iou_sum):
  """Bounding-boxes self-suppression loop body.

  Args:
    iou: A float Tensor with shape [1, num_boxes, max_num_instance]: IOUs.
    iou_threshold: A scalar, representing IOU threshold.
    loop_condition: The loop condition returned from last iteration.
    iou_sum: iou_sum_new returned from last iteration.

  Returns:
    iou_suppressed: A float Tensor with shape [1, num_boxes, max_num_instance],
                    IOU after suppression.
    iou_threshold: A scalar, representing IOU threshold.
    loop_condition: Bool Tensor of shape [], the loop condition.
    iou_sum_new: The new IOU sum.
  """
  del loop_condition
  can_suppress_others = tf.cast(
      tf.reshape(tf.reduce_max(iou, 1) <= iou_threshold, [1, -1, 1]), iou.dtype)
  iou_suppressed = tf.reshape(
      tf.cast(
          tf.reduce_max(can_suppress_others * iou, 1) <= iou_threshold,
          iou.dtype), [1, -1, 1]) * iou
  iou_sum_new = tf.reduce_sum(iou_suppressed, [1, 2])
  return [
      iou_suppressed, iou_threshold,
      tf.reduce_any(iou_sum - iou_sum_new > iou_threshold), iou_sum_new
  ]


def _cross_suppression(boxes, box_slice, iou_threshold, inner_idx):
  """Bounding-boxes cross-suppression loop body.

  Args:
    boxes: A float Tensor of shape [1, anchors, 4], representing boxes.
    box_slice: A float Tensor of shape [1, _NMS_TILE_SIZE, 4], the box tile
      returned from last iteration
    iou_threshold: A scalar, representing IOU threshold.
    inner_idx: A scalar, representing inner index.

  Returns:
    boxes: A float Tensor of shape [1, anchors, 4], representing boxes.
    ret_slice: A float Tensor of shape [1, _NMS_TILE_SIZE, 4], the box tile
               after suppression
    iou_threshold: A scalar, representing IOU threshold.
    inner_idx: A scalar, inner index incremented.
  """
  new_slice = tf.slice(boxes, [0, inner_idx * _NMS_TILE_SIZE, 0],
                       [1, _NMS_TILE_SIZE, 4])
  iou = batch_iou(new_slice, box_slice)
  ret_slice = tf.expand_dims(
      tf.cast(tf.reduce_all(iou < iou_threshold, [1]), box_slice.dtype),
      2) * box_slice
  return boxes, ret_slice, iou_threshold, inner_idx + 1


def _suppression_loop_body(boxes, iou_threshold, output_size, idx):
  """Process boxes in the range [idx*_NMS_TILE_SIZE, (idx+1)*_NMS_TILE_SIZE).

  Args:
    boxes: a tensor with a shape of [1, anchors, 4].
    iou_threshold: a float representing the threshold for deciding whether boxes
      overlap too much with respect to IOU.
    output_size: an int32 tensor of size [1]. Representing the number of
      selected boxes.
    idx: an integer scalar representing induction variable.

  Returns:
    boxes: updated boxes.
    iou_threshold: pass down iou_threshold to the next iteration.
    output_size: the updated output_size.
    idx: the updated induction variable.
  """
  num_tiles = tf.shape(boxes)[1] // _NMS_TILE_SIZE

  # Iterates over tiles that can possibly suppress the current tile.
  box_slice = tf.slice(boxes, [0, idx * _NMS_TILE_SIZE, 0],
                       [1, _NMS_TILE_SIZE, 4])
  _, box_slice, _, _ = tf.while_loop(
      lambda _boxes, _box_slice, _threshold, inner_idx: inner_idx < idx,
      _cross_suppression, [boxes, box_slice, iou_threshold,
                           tf.constant(0)])

  # Iterates over the current tile to compute self-suppression.
  iou = batch_iou(box_slice, box_slice)
  mask = tf.expand_dims(
      tf.reshape(tf.range(_NMS_TILE_SIZE), [1, -1]) > tf.reshape(
          tf.range(_NMS_TILE_SIZE), [-1, 1]), 0)
  iou *= tf.cast(tf.logical_and(mask, iou >= iou_threshold), iou.dtype)
  suppressed_iou, _, _, _ = tf.while_loop(
      lambda _iou, _threshold, loop_condition, _iou_sum: loop_condition,
      _self_suppression,
      [iou, iou_threshold,
       tf.constant(True),
       tf.reduce_sum(iou, [1, 2])])
  suppressed_box = tf.reduce_sum(suppressed_iou, 1) > 0
  box_slice *= tf.expand_dims(1.0 - tf.cast(suppressed_box, box_slice.dtype), 2)

  # Uses box_slice to update the input boxes.
  mask = tf.reshape(
      tf.cast(tf.equal(tf.range(num_tiles), idx), boxes.dtype), [1, -1, 1, 1])
  boxes = tf.tile(tf.expand_dims(box_slice, [1]),
                  [1, num_tiles, 1, 1]) * mask + tf.reshape(
                      boxes, [1, num_tiles, _NMS_TILE_SIZE, 4]) * (1 - mask)
  boxes = tf.reshape(boxes, [1, -1, 4])

  # Updates output_size.
  output_size += tf.reduce_sum(
      tf.cast(tf.reduce_any(box_slice > 0, [2]), tf.int32), [1])
  return boxes, iou_threshold, output_size, idx + 1


def partitioned_non_max_suppression_padded(boxes,
                                           scores,
                                           max_output_size,
                                           iou_threshold=0.5,
                                           score_threshold=float('-inf')):
  """A tiled version of [`tf.image.non_max_suppression_padded`](https://www.tensorflow.org/api_docs/python/tf/image/non_max_suppression_padded).

  The overall design of the algorithm is to handle boxes tile-by-tile:

  boxes = boxes.pad_to_multiple_of(tile_size)
  num_tiles = len(boxes) // tile_size
  output_boxes = []
  for i in range(num_tiles):
    box_tile = boxes[i*tile_size : (i+1)*tile_size]
    for j in range(i - 1):
      suppressing_tile = boxes[j*tile_size : (j+1)*tile_size]
      iou = batch_iou(box_tile, suppressing_tile)
      # if the box is suppressed in iou, clear it to a dot
      box_tile *= _update_boxes(iou)
    # Iteratively handle the diagonal tile.
    iou = _box_overlap(box_tile, box_tile)
    iou_changed = True
    while iou_changed:
      # boxes that are not suppressed by anything else
      suppressing_boxes = _get_suppressing_boxes(iou)
      # boxes that are suppressed by suppressing_boxes
      suppressed_boxes = _get_suppressed_boxes(iou, suppressing_boxes)
      # clear iou to 0 for boxes that are suppressed, as they cannot be used
      # to suppress other boxes any more
      new_iou = _clear_iou(iou, suppressed_boxes)
      iou_changed = (new_iou != iou)
      iou = new_iou
    # remaining boxes that can still suppress others, are selected boxes.
    output_boxes.append(_get_suppressing_boxes(iou))
    if len(output_boxes) >= max_output_size:
      break

  Args:
    boxes: A 2-D float `Tensor` of shape `[num_boxes, 4]`.
    scores: A 1-D float `Tensor` of shape `[num_boxes]` representing a single
      score corresponding to each box (each row of boxes).
    max_output_size: a scalar integer `Tensor` representing the maximum number
      of boxes to be selected by non max suppression.
    iou_threshold: a float representing the threshold for deciding whether boxes
      overlap too much with respect to IOU.
    score_threshold: A float representing the threshold for deciding when to
      remove boxes based on score.

  Returns:
    selected_indices: a tensor of shape [anchors].
    num_valid_boxes: a scalar int tensor.
    nms_proposals: a tensor with a shape of [anchors, 4]. It has
      same dtype as input boxes.
    nms_scores: a tensor with a shape of [anchors]. It has same
      dtype as input scores.
    argsort_ids: a tensor of shape [anchors], mapping from input order of boxes
      to output order of boxes.
  """
  num_boxes = tf.shape(boxes)[0]
  pad = tf.cast(
      tf.ceil(tf.cast(num_boxes, tf.float32) / _NMS_TILE_SIZE),
      tf.int32) * _NMS_TILE_SIZE - num_boxes

  scores, argsort_ids = tf.nn.top_k(scores, k=num_boxes, sorted=True)
  boxes = tf.gather(boxes, argsort_ids)
  num_boxes = tf.shape(boxes)[0]
  num_boxes += pad
  boxes = tf.pad(
      tf.cast(boxes, tf.float32), [[0, pad], [0, 0]], constant_values=-1)
  scores = tf.pad(tf.cast(scores, tf.float32), [[0, pad]])

  # mask boxes to -1 by score threshold
  scores_mask = tf.expand_dims(
      tf.cast(scores > score_threshold, boxes.dtype), axis=1)
  boxes = ((boxes + 1.) * scores_mask) - 1.

  boxes = tf.expand_dims(boxes, axis=0)
  scores = tf.expand_dims(scores, axis=0)

  def _loop_cond(unused_boxes, unused_threshold, output_size, idx):
    return tf.logical_and(
        tf.reduce_min(output_size) < max_output_size,
        idx < num_boxes // _NMS_TILE_SIZE)

  selected_boxes, _, output_size, _ = tf.while_loop(
      _loop_cond, _suppression_loop_body,
      [boxes, iou_threshold,
       tf.zeros([1], tf.int32),
       tf.constant(0)])
  idx = num_boxes - tf.cast(
      tf.nn.top_k(
          tf.cast(tf.reduce_any(selected_boxes > 0, [2]), tf.int32) *
          tf.expand_dims(tf.range(num_boxes, 0, -1), 0), max_output_size)[0],
      tf.int32)
  idx = tf.minimum(idx, num_boxes - 1 - pad)
  idx = tf.reshape(idx + tf.reshape(tf.range(1) * num_boxes, [-1, 1]), [-1])
  num_valid_boxes = tf.reduce_sum(output_size)
  return (idx, num_valid_boxes, tf.reshape(boxes, [-1, 4]),
          tf.reshape(scores, [-1]), argsort_ids)


def _validate_boxes_scores_iou_thresh(boxes, scores, iou_thresh,
                                      change_coordinate_frame, clip_window):
  """Validates boxes, scores and iou_thresh.

  This function validates the boxes, scores, iou_thresh
     and if change_coordinate_frame is True, clip_window must be specified.

  Args:
    boxes: A [k, q, 4] float32 tensor containing k detections. `q` can be either
      number of classes or 1 depending on whether a separate box is predicted
      per class.
    scores: A [k, num_classes] float32 tensor containing the scores for each of
      the k detections. The scores have to be non-negative when
      pad_to_max_output_size is True.
    iou_thresh: scalar threshold for IOU (new boxes that have high IOU overlap
      with previously selected boxes are removed).
    change_coordinate_frame: Whether to normalize coordinates after clipping
      relative to clip_window (this can only be set to True if a clip_window is
      provided)
    clip_window: A float32 tensor of the form [y_min, x_min, y_max, x_max]
      representing the window to clip and normalize boxes to before performing
      non-max suppression.

  Raises:
    ValueError: if iou_thresh is not in [0, 1] or if input boxlist does not
    have a valid scores field.
  """
  if not 0 <= iou_thresh <= 1.0:
    raise ValueError('iou_thresh must be between 0 and 1')
  if scores.shape.ndims != 2:
    raise ValueError('scores field must be of rank 2')
  if shape_utils.get_dim_as_int(scores.shape[1]) is None:
    raise ValueError('scores must have statically defined second ' 'dimension')
  if boxes.shape.ndims != 3:
    raise ValueError('boxes must be of rank 3.')
  if not (shape_utils.get_dim_as_int(
      boxes.shape[1]) == shape_utils.get_dim_as_int(scores.shape[1]) or
          shape_utils.get_dim_as_int(boxes.shape[1]) == 1):
    raise ValueError('second dimension of boxes must be either 1 or equal '
                     'to the second dimension of scores')
  if shape_utils.get_dim_as_int(boxes.shape[2]) != 4:
    raise ValueError('last dimension of boxes must be of size 4.')
  if change_coordinate_frame and clip_window is None:
    raise ValueError('if change_coordinate_frame is True, then a clip_window'
                     'must be specified.')


def _clip_window_prune_boxes(sorted_boxes, clip_window, pad_to_max_output_size,
                             change_coordinate_frame):
  """Prune boxes with zero area.

  Args:
    sorted_boxes: A BoxList containing k detections.
    clip_window: A float32 tensor of the form [y_min, x_min, y_max, x_max]
      representing the window to clip and normalize boxes to before performing
      non-max suppression.
    pad_to_max_output_size: flag indicating whether to pad to max output size or
      not.
    change_coordinate_frame: Whether to normalize coordinates after clipping
      relative to clip_window (this can only be set to True if a clip_window is
      provided).

  Returns:
    sorted_boxes: A BoxList containing k detections after pruning.
    num_valid_nms_boxes_cumulative: Number of valid NMS boxes
  """
  sorted_boxes = box_list_ops.clip_to_window(
      sorted_boxes,
      clip_window,
      filter_nonoverlapping=not pad_to_max_output_size)
  # Set the scores of boxes with zero area to -1 to keep the default
  # behaviour of pruning out zero area boxes.
  sorted_boxes_size = tf.shape(sorted_boxes.get())[0]
  non_zero_box_area = tf.cast(box_list_ops.area(sorted_boxes), tf.bool)
  sorted_boxes_scores = tf.where(
      non_zero_box_area, sorted_boxes.get_field(fields.BoxListFields.scores),
      -1 * tf.ones(sorted_boxes_size))
  sorted_boxes.add_field(fields.BoxListFields.scores, sorted_boxes_scores)
  num_valid_nms_boxes_cumulative = tf.reduce_sum(
      tf.cast(tf.greater_equal(sorted_boxes_scores, 0), tf.int32))
  sorted_boxes = box_list_ops.sort_by_field(sorted_boxes,
                                            fields.BoxListFields.scores)
  if change_coordinate_frame:
    sorted_boxes = box_list_ops.change_coordinate_frame(sorted_boxes,
                                                        clip_window)
    if sorted_boxes.has_field(fields.BoxListFields.keypoints):
      sorted_keypoints = sorted_boxes.get_field(fields.BoxListFields.keypoints)
      sorted_keypoints = keypoint_ops.change_coordinate_frame(sorted_keypoints,
                                                              clip_window)
      sorted_boxes.set_field(fields.BoxListFields.keypoints, sorted_keypoints)
  return sorted_boxes, num_valid_nms_boxes_cumulative


class NullContextmanager(object):

  def __enter__(self):
    pass

  def __exit__(self, type_arg, value_arg, traceback_arg):
    return False


def multiclass_non_max_suppression(boxes,
                                   scores,
                                   score_thresh,
                                   iou_thresh,
                                   max_size_per_class,
                                   max_total_size=0,
                                   clip_window=None,
                                   change_coordinate_frame=False,
                                   masks=None,
                                   boundaries=None,
                                   pad_to_max_output_size=False,
                                   use_partitioned_nms=False,
                                   additional_fields=None,
                                   soft_nms_sigma=0.0,
                                   use_hard_nms=False,
                                   use_cpu_nms=False,
                                   scope=None):
  """Multi-class version of non maximum suppression.

  This op greedily selects a subset of detection bounding boxes, pruning
  away boxes that have high IOU (intersection over union) overlap (> thresh)
  with already selected boxes.  It operates independently for each class for
  which scores are provided (via the scores field of the input box_list),
  pruning boxes with score less than a provided threshold prior to
  applying NMS.

  Please note that this operation is performed on *all* classes, therefore any
  background classes should be removed prior to calling this function.

  Selected boxes are guaranteed to be sorted in decreasing order by score (but
  the sort is not guaranteed to be stable).

  Args:
    boxes: A [k, q, 4] float32 tensor containing k detections. `q` can be either
      number of classes or 1 depending on whether a separate box is predicted
      per class.
    scores: A [k, num_classes] float32 tensor containing the scores for each of
      the k detections. The scores have to be non-negative when
      pad_to_max_output_size is True.
    score_thresh: scalar threshold for score (low scoring boxes are removed).
    iou_thresh: scalar threshold for IOU (new boxes that have high IOU overlap
      with previously selected boxes are removed).
    max_size_per_class: maximum number of retained boxes per class.
    max_total_size: maximum number of boxes retained over all classes. By
      default returns all boxes retained after capping boxes per class.
    clip_window: A float32 tensor of the form [y_min, x_min, y_max, x_max]
      representing the window to clip and normalize boxes to before performing
      non-max suppression.
    change_coordinate_frame: Whether to normalize coordinates after clipping
      relative to clip_window (this can only be set to True if a clip_window
      is provided)
    masks: (optional) a [k, q, mask_height, mask_width] float32 tensor
      containing box masks. `q` can be either number of classes or 1 depending
      on whether a separate mask is predicted per class.
    boundaries: (optional) a [k, q, boundary_height, boundary_width] float32
      tensor containing box boundaries. `q` can be either number of classes or 1
      depending on whether a separate boundary is predicted per class.
    pad_to_max_output_size: If true, the output nmsed boxes are padded to be of
      length `max_size_per_class`. Defaults to false.
    use_partitioned_nms: If true, use partitioned version of
      non_max_suppression.
    additional_fields: (optional) If not None, a dictionary that maps keys to
      tensors whose first dimensions are all of size `k`. After non-maximum
      suppression, all tensors corresponding to the selected boxes will be
      added to resulting BoxList.
    soft_nms_sigma: A scalar float representing the Soft NMS sigma parameter;
      See Bodla et al, https://arxiv.org/abs/1704.04503).  When
      `soft_nms_sigma=0.0` (which is default), we fall back to standard (hard)
      NMS.  Soft NMS is currently only supported when pad_to_max_output_size is
      False.
    use_hard_nms: Enforce the usage of hard NMS.
    use_cpu_nms: Enforce NMS to run on CPU.
    scope: name scope.

  Returns:
    A tuple of sorted_boxes and num_valid_nms_boxes. The sorted_boxes is a
      BoxList holds M boxes with a rank-1 scores field representing
      corresponding scores for each box with scores sorted in decreasing order
      and a rank-1 classes field representing a class label for each box. The
      num_valid_nms_boxes is a 0-D integer tensor representing the number of
      valid elements in `BoxList`, with the valid elements appearing first.

  Raises:
    ValueError: if iou_thresh is not in [0, 1] or if input boxlist does not have
      a valid scores field.
    ValueError: if Soft NMS (tf.image.non_max_suppression_with_scores) is not
      supported in the current TF version and `soft_nms_sigma` is nonzero.
  """
  _validate_boxes_scores_iou_thresh(boxes, scores, iou_thresh,
                                    change_coordinate_frame, clip_window)
  if pad_to_max_output_size and soft_nms_sigma != 0.0:
    raise ValueError('Soft NMS (soft_nms_sigma != 0.0) is currently not '
                     'supported when pad_to_max_output_size is True.')

  with tf.name_scope(scope, 'MultiClassNonMaxSuppression'), tf.device(
      'cpu:0') if use_cpu_nms else NullContextmanager():
    num_scores = tf.shape(scores)[0]
    num_classes = shape_utils.get_dim_as_int(scores.get_shape()[1])

    selected_boxes_list = []
    num_valid_nms_boxes_cumulative = tf.constant(0)
    per_class_boxes_list = tf.unstack(boxes, axis=1)
    if masks is not None:
      per_class_masks_list = tf.unstack(masks, axis=1)
    if boundaries is not None:
      per_class_boundaries_list = tf.unstack(boundaries, axis=1)
    boxes_ids = (range(num_classes) if len(per_class_boxes_list) > 1
                 else [0] * num_classes)
    for class_idx, boxes_idx in zip(range(num_classes), boxes_ids):
      per_class_boxes = per_class_boxes_list[boxes_idx]
      boxlist_and_class_scores = box_list.BoxList(per_class_boxes)
      class_scores = tf.reshape(
          tf.slice(scores, [0, class_idx], tf.stack([num_scores, 1])), [-1])

      boxlist_and_class_scores.add_field(fields.BoxListFields.scores,
                                         class_scores)
      if masks is not None:
        per_class_masks = per_class_masks_list[boxes_idx]
        boxlist_and_class_scores.add_field(fields.BoxListFields.masks,
                                           per_class_masks)
      if boundaries is not None:
        per_class_boundaries = per_class_boundaries_list[boxes_idx]
        boxlist_and_class_scores.add_field(fields.BoxListFields.boundaries,
                                           per_class_boundaries)
      if additional_fields is not None:
        for key, tensor in additional_fields.items():
          boxlist_and_class_scores.add_field(key, tensor)

      nms_result = None
      selected_scores = None
      if pad_to_max_output_size:
        max_selection_size = max_size_per_class
        if use_partitioned_nms:
          (selected_indices, num_valid_nms_boxes,
           boxlist_and_class_scores.data['boxes'],
           boxlist_and_class_scores.data['scores'],
           _) = partitioned_non_max_suppression_padded(
               boxlist_and_class_scores.get(),
               boxlist_and_class_scores.get_field(fields.BoxListFields.scores),
               max_selection_size,
               iou_threshold=iou_thresh,
               score_threshold=score_thresh)
        else:
          selected_indices, num_valid_nms_boxes = (
              tf.image.non_max_suppression_padded(
                  boxlist_and_class_scores.get(),
                  boxlist_and_class_scores.get_field(
                      fields.BoxListFields.scores),
                  max_selection_size,
                  iou_threshold=iou_thresh,
                  score_threshold=score_thresh,
                  pad_to_max_output_size=True))
        nms_result = box_list_ops.gather(boxlist_and_class_scores,
                                         selected_indices)
        selected_scores = nms_result.get_field(fields.BoxListFields.scores)
      else:
        max_selection_size = tf.minimum(max_size_per_class,
                                        boxlist_and_class_scores.num_boxes())
        if (hasattr(tf.image, 'non_max_suppression_with_scores') and
            tf.compat.forward_compatible(2019, 6, 6) and not use_hard_nms):
          (selected_indices, selected_scores
          ) = tf.image.non_max_suppression_with_scores(
              boxlist_and_class_scores.get(),
              boxlist_and_class_scores.get_field(fields.BoxListFields.scores),
              max_selection_size,
              iou_threshold=iou_thresh,
              score_threshold=score_thresh,
              soft_nms_sigma=soft_nms_sigma)
          num_valid_nms_boxes = tf.shape(selected_indices)[0]
          selected_indices = tf.concat(
              [selected_indices,
               tf.zeros(max_selection_size-num_valid_nms_boxes, tf.int32)], 0)
          selected_scores = tf.concat(
              [selected_scores,
               tf.zeros(max_selection_size-num_valid_nms_boxes,
                        tf.float32)], -1)
          nms_result = box_list_ops.gather(boxlist_and_class_scores,
                                           selected_indices)
        else:
          if soft_nms_sigma != 0:
            raise ValueError('Soft NMS not supported in current TF version!')
          selected_indices = tf.image.non_max_suppression(
              boxlist_and_class_scores.get(),
              boxlist_and_class_scores.get_field(fields.BoxListFields.scores),
              max_selection_size,
              iou_threshold=iou_thresh,
              score_threshold=score_thresh)
          num_valid_nms_boxes = tf.shape(selected_indices)[0]
          selected_indices = tf.concat(
              [selected_indices,
               tf.zeros(max_selection_size-num_valid_nms_boxes, tf.int32)], 0)
          nms_result = box_list_ops.gather(boxlist_and_class_scores,
                                           selected_indices)
          selected_scores = nms_result.get_field(fields.BoxListFields.scores)
      # Make the scores -1 for invalid boxes.
      valid_nms_boxes_indices = tf.less(
          tf.range(max_selection_size), num_valid_nms_boxes)

      nms_result.add_field(
          fields.BoxListFields.scores,
          tf.where(valid_nms_boxes_indices,
                   selected_scores, -1*tf.ones(max_selection_size)))
      num_valid_nms_boxes_cumulative += num_valid_nms_boxes

      nms_result.add_field(
          fields.BoxListFields.classes, (tf.zeros_like(
              nms_result.get_field(fields.BoxListFields.scores)) + class_idx))
      selected_boxes_list.append(nms_result)
    selected_boxes = box_list_ops.concatenate(selected_boxes_list)
    sorted_boxes = box_list_ops.sort_by_field(selected_boxes,
                                              fields.BoxListFields.scores)
    if clip_window is not None:
      # When pad_to_max_output_size is False, it prunes the boxes with zero
      # area.
      sorted_boxes, num_valid_nms_boxes_cumulative = _clip_window_prune_boxes(
          sorted_boxes, clip_window, pad_to_max_output_size,
          change_coordinate_frame)

    if max_total_size:
      max_total_size = tf.minimum(max_total_size, sorted_boxes.num_boxes())
      sorted_boxes = box_list_ops.gather(sorted_boxes, tf.range(max_total_size))
      num_valid_nms_boxes_cumulative = tf.where(
          max_total_size > num_valid_nms_boxes_cumulative,
          num_valid_nms_boxes_cumulative, max_total_size)
    # Select only the valid boxes if pad_to_max_output_size is False.
    if not pad_to_max_output_size:
      sorted_boxes = box_list_ops.gather(
          sorted_boxes, tf.range(num_valid_nms_boxes_cumulative))

    return sorted_boxes, num_valid_nms_boxes_cumulative


def class_agnostic_non_max_suppression(boxes,
                                       scores,
                                       score_thresh,
                                       iou_thresh,
                                       max_classes_per_detection=1,
                                       max_total_size=0,
                                       clip_window=None,
                                       change_coordinate_frame=False,
                                       masks=None,
                                       boundaries=None,
                                       pad_to_max_output_size=False,
                                       use_partitioned_nms=False,
                                       additional_fields=None,
                                       soft_nms_sigma=0.0,
                                       scope=None):
  """Class-agnostic version of non maximum suppression.

  This op greedily selects a subset of detection bounding boxes, pruning
  away boxes that have high IOU (intersection over union) overlap (> thresh)
  with already selected boxes.  It operates on all the boxes using
  max scores across all classes for which scores are provided (via the scores
  field of the input box_list), pruning boxes with score less than a provided
  threshold prior to applying NMS.

  Please note that this operation is performed in a class-agnostic way,
  therefore any background classes should be removed prior to calling this
  function.

  Selected boxes are guaranteed to be sorted in decreasing order by score (but
  the sort is not guaranteed to be stable).

  Args:
    boxes: A [k, q, 4] float32 tensor containing k detections. `q` can be either
      number of classes or 1 depending on whether a separate box is predicted
      per class.
    scores: A [k, num_classes] float32 tensor containing the scores for each of
      the k detections. The scores have to be non-negative when
      pad_to_max_output_size is True.
    score_thresh: scalar threshold for score (low scoring boxes are removed).
    iou_thresh: scalar threshold for IOU (new boxes that have high IOU overlap
      with previously selected boxes are removed).
    max_classes_per_detection: maximum number of retained classes per detection
      box in class-agnostic NMS.
    max_total_size: maximum number of boxes retained over all classes. By
      default returns all boxes retained after capping boxes per class.
    clip_window: A float32 tensor of the form [y_min, x_min, y_max, x_max]
      representing the window to clip and normalize boxes to before performing
      non-max suppression.
    change_coordinate_frame: Whether to normalize coordinates after clipping
      relative to clip_window (this can only be set to True if a clip_window is
      provided)
    masks: (optional) a [k, q, mask_height, mask_width] float32 tensor
      containing box masks. `q` can be either number of classes or 1 depending
      on whether a separate mask is predicted per class.
    boundaries: (optional) a [k, q, boundary_height, boundary_width] float32
      tensor containing box boundaries. `q` can be either number of classes or 1
      depending on whether a separate boundary is predicted per class.
    pad_to_max_output_size: If true, the output nmsed boxes are padded to be of
      length `max_size_per_class`. Defaults to false.
    use_partitioned_nms: If true, use partitioned version of
      non_max_suppression.
    additional_fields: (optional) If not None, a dictionary that maps keys to
      tensors whose first dimensions are all of size `k`. After non-maximum
      suppression, all tensors corresponding to the selected boxes will be added
      to resulting BoxList.
    soft_nms_sigma: A scalar float representing the Soft NMS sigma parameter;
      See Bodla et al, https://arxiv.org/abs/1704.04503).  When
      `soft_nms_sigma=0.0` (which is default), we fall back to standard (hard)
      NMS.  Soft NMS is currently only supported when pad_to_max_output_size is
      False.
    scope: name scope.

  Returns:
    A tuple of sorted_boxes and num_valid_nms_boxes. The sorted_boxes is a
      BoxList holds M boxes with a rank-1 scores field representing
      corresponding scores for each box with scores sorted in decreasing order
      and a rank-1 classes field representing a class label for each box. The
      num_valid_nms_boxes is a 0-D integer tensor representing the number of
      valid elements in `BoxList`, with the valid elements appearing first.

  Raises:
    ValueError: if iou_thresh is not in [0, 1] or if input boxlist does not have
      a valid scores field or if non-zero soft_nms_sigma is provided when
      pad_to_max_output_size is True.
  """
  _validate_boxes_scores_iou_thresh(boxes, scores, iou_thresh,
                                    change_coordinate_frame, clip_window)
  if pad_to_max_output_size and soft_nms_sigma != 0.0:
    raise ValueError('Soft NMS (soft_nms_sigma != 0.0) is currently not '
                     'supported when pad_to_max_output_size is True.')

  if max_classes_per_detection > 1:
    raise ValueError('Max classes per detection box >1 not supported.')
  q = shape_utils.get_dim_as_int(boxes.shape[1])
  if q > 1:
    class_ids = tf.expand_dims(
        tf.argmax(scores, axis=1, output_type=tf.int32), axis=1)
    boxes = tf.batch_gather(boxes, class_ids)
    if masks is not None:
      masks = tf.batch_gather(masks, class_ids)
    if boundaries is not None:
      boundaries = tf.batch_gather(boundaries, class_ids)
  boxes = tf.squeeze(boxes, axis=[1])
  if masks is not None:
    masks = tf.squeeze(masks, axis=[1])
  if boundaries is not None:
    boundaries = tf.squeeze(boundaries, axis=[1])

  with tf.name_scope(scope, 'ClassAgnosticNonMaxSuppression'):
    boxlist_and_class_scores = box_list.BoxList(boxes)
    max_scores = tf.reduce_max(scores, axis=-1)
    classes_with_max_scores = tf.argmax(scores, axis=-1)
    boxlist_and_class_scores.add_field(fields.BoxListFields.scores, max_scores)
    if masks is not None:
      boxlist_and_class_scores.add_field(fields.BoxListFields.masks, masks)
    if boundaries is not None:
      boxlist_and_class_scores.add_field(fields.BoxListFields.boundaries,
                                         boundaries)

    if additional_fields is not None:
      for key, tensor in additional_fields.items():
        boxlist_and_class_scores.add_field(key, tensor)

    nms_result = None
    selected_scores = None
    if pad_to_max_output_size:
      max_selection_size = max_total_size
      if use_partitioned_nms:
        (selected_indices, num_valid_nms_boxes,
         boxlist_and_class_scores.data['boxes'],
         boxlist_and_class_scores.data['scores'],
         argsort_ids) = partitioned_non_max_suppression_padded(
             boxlist_and_class_scores.get(),
             boxlist_and_class_scores.get_field(fields.BoxListFields.scores),
             max_selection_size,
             iou_threshold=iou_thresh,
             score_threshold=score_thresh)
        classes_with_max_scores = tf.gather(classes_with_max_scores,
                                            argsort_ids)
      else:
        selected_indices, num_valid_nms_boxes = (
            tf.image.non_max_suppression_padded(
                boxlist_and_class_scores.get(),
                boxlist_and_class_scores.get_field(fields.BoxListFields.scores),
                max_selection_size,
                iou_threshold=iou_thresh,
                score_threshold=score_thresh,
                pad_to_max_output_size=True))
      nms_result = box_list_ops.gather(boxlist_and_class_scores,
                                       selected_indices)
      selected_scores = nms_result.get_field(fields.BoxListFields.scores)
    else:
      max_selection_size = tf.minimum(max_total_size,
                                      boxlist_and_class_scores.num_boxes())
      if (hasattr(tf.image, 'non_max_suppression_with_scores') and
          tf.compat.forward_compatible(2019, 6, 6)):
        (selected_indices, selected_scores
        ) = tf.image.non_max_suppression_with_scores(
            boxlist_and_class_scores.get(),
            boxlist_and_class_scores.get_field(fields.BoxListFields.scores),
            max_selection_size,
            iou_threshold=iou_thresh,
            score_threshold=score_thresh,
            soft_nms_sigma=soft_nms_sigma)
        num_valid_nms_boxes = tf.shape(selected_indices)[0]
        selected_indices = tf.concat([
            selected_indices,
            tf.zeros(max_selection_size - num_valid_nms_boxes, tf.int32)
        ], 0)
        selected_scores = tf.concat(
            [selected_scores,
             tf.zeros(max_selection_size-num_valid_nms_boxes, tf.float32)], -1)
        nms_result = box_list_ops.gather(boxlist_and_class_scores,
                                         selected_indices)
      else:
        if soft_nms_sigma != 0:
          raise ValueError('Soft NMS not supported in current TF version!')
        selected_indices = tf.image.non_max_suppression(
            boxlist_and_class_scores.get(),
            boxlist_and_class_scores.get_field(fields.BoxListFields.scores),
            max_selection_size,
            iou_threshold=iou_thresh,
            score_threshold=score_thresh)
        num_valid_nms_boxes = tf.shape(selected_indices)[0]
        selected_indices = tf.concat(
            [selected_indices,
             tf.zeros(max_selection_size-num_valid_nms_boxes, tf.int32)], 0)
        nms_result = box_list_ops.gather(boxlist_and_class_scores,
                                         selected_indices)
        selected_scores = nms_result.get_field(fields.BoxListFields.scores)
    valid_nms_boxes_indices = tf.less(
        tf.range(max_selection_size), num_valid_nms_boxes)
    nms_result.add_field(
        fields.BoxListFields.scores,
        tf.where(valid_nms_boxes_indices,
                 selected_scores, -1*tf.ones(max_selection_size)))

    selected_classes = tf.gather(classes_with_max_scores, selected_indices)
    selected_classes = tf.cast(selected_classes, tf.float32)
    nms_result.add_field(fields.BoxListFields.classes, selected_classes)
    selected_boxes = nms_result
    sorted_boxes = box_list_ops.sort_by_field(selected_boxes,
                                              fields.BoxListFields.scores)

    if clip_window is not None:
      # When pad_to_max_output_size is False, it prunes the boxes with zero
      # area.
      sorted_boxes, num_valid_nms_boxes = _clip_window_prune_boxes(
          sorted_boxes, clip_window, pad_to_max_output_size,
          change_coordinate_frame)

    if max_total_size:
      max_total_size = tf.minimum(max_total_size, sorted_boxes.num_boxes())
      sorted_boxes = box_list_ops.gather(sorted_boxes, tf.range(max_total_size))
      num_valid_nms_boxes = tf.where(max_total_size > num_valid_nms_boxes,
                                     num_valid_nms_boxes, max_total_size)
    # Select only the valid boxes if pad_to_max_output_size is False.
    if not pad_to_max_output_size:
      sorted_boxes = box_list_ops.gather(sorted_boxes,
                                         tf.range(num_valid_nms_boxes))

    return sorted_boxes, num_valid_nms_boxes


def batch_multiclass_non_max_suppression(boxes,
                                         scores,
                                         score_thresh,
                                         iou_thresh,
                                         max_size_per_class,
                                         max_total_size=0,
                                         clip_window=None,
                                         change_coordinate_frame=False,
                                         num_valid_boxes=None,
                                         masks=None,
                                         additional_fields=None,
                                         soft_nms_sigma=0.0,
                                         scope=None,
                                         use_static_shapes=False,
                                         use_partitioned_nms=False,
                                         parallel_iterations=32,
                                         use_class_agnostic_nms=False,
                                         max_classes_per_detection=1,
                                         use_dynamic_map_fn=False,
                                         use_combined_nms=False,
                                         use_hard_nms=False,
                                         use_cpu_nms=False):
  """Multi-class version of non maximum suppression that operates on a batch.

  This op is similar to `multiclass_non_max_suppression` but operates on a batch
  of boxes and scores. See documentation for `multiclass_non_max_suppression`
  for details.

  Args:
    boxes: A [batch_size, num_anchors, q, 4] float32 tensor containing
      detections. If `q` is 1 then same boxes are used for all classes
      otherwise, if `q` is equal to number of classes, class-specific boxes are
      used.
    scores: A [batch_size, num_anchors, num_classes] float32 tensor containing
      the scores for each of the `num_anchors` detections. The scores have to be
      non-negative when use_static_shapes is set True.
    score_thresh: scalar threshold for score (low scoring boxes are removed).
    iou_thresh: scalar threshold for IOU (new boxes that have high IOU overlap
      with previously selected boxes are removed).
    max_size_per_class: maximum number of retained boxes per class.
    max_total_size: maximum number of boxes retained over all classes. By
      default returns all boxes retained after capping boxes per class.
    clip_window: A float32 tensor of shape [batch_size, 4]  where each entry is
      of the form [y_min, x_min, y_max, x_max] representing the window to clip
      boxes to before performing non-max suppression. This argument can also be
      a tensor of shape [4] in which case, the same clip window is applied to
      all images in the batch. If clip_widow is None, all boxes are used to
      perform non-max suppression.
    change_coordinate_frame: Whether to normalize coordinates after clipping
      relative to clip_window (this can only be set to True if a clip_window is
      provided)
    num_valid_boxes: (optional) a Tensor of type `int32`. A 1-D tensor of shape
      [batch_size] representing the number of valid boxes to be considered for
      each image in the batch.  This parameter allows for ignoring zero
      paddings.
    masks: (optional) a [batch_size, num_anchors, q, mask_height, mask_width]
      float32 tensor containing box masks. `q` can be either number of classes
      or 1 depending on whether a separate mask is predicted per class.
    additional_fields: (optional) If not None, a dictionary that maps keys to
      tensors whose dimensions are [batch_size, num_anchors, ...].
    soft_nms_sigma: A scalar float representing the Soft NMS sigma parameter;
      See Bodla et al, https://arxiv.org/abs/1704.04503).  When
      `soft_nms_sigma=0.0` (which is default), we fall back to standard (hard)
      NMS.  Soft NMS is currently only supported when pad_to_max_output_size is
      False.
    scope: tf scope name.
    use_static_shapes: If true, the output nmsed boxes are padded to be of
      length `max_size_per_class` and it doesn't clip boxes to max_total_size.
      Defaults to false.
    use_partitioned_nms: If true, use partitioned version of
      non_max_suppression.
    parallel_iterations: (optional) number of batch items to process in
      parallel.
    use_class_agnostic_nms: If true, this uses class-agnostic non max
      suppression
    max_classes_per_detection: Maximum number of retained classes per detection
      box in class-agnostic NMS.
    use_dynamic_map_fn: If true, images in the batch will be processed within a
      dynamic loop. Otherwise, a static loop will be used if possible.
    use_combined_nms: If true, it uses tf.image.combined_non_max_suppression (
      multi-class version of NMS that operates on a batch).
      It greedily selects a subset of detection bounding boxes, pruning away
      boxes that have high IOU (intersection over union) overlap (> thresh) with
      already selected boxes. It operates independently for each batch.
      Within each batch, it operates independently for each class for which
      scores are provided (via the scores field of the input box_list),
      pruning boxes with score less than a provided threshold prior to applying
      NMS. This operation is performed on *all* batches and *all* classes
      in the batch, therefore any background classes should be removed prior to
      calling this function.
      Masks and additional fields are not supported.
      See argument checks in the code below for unsupported arguments.
    use_hard_nms: Enforce the usage of hard NMS.
    use_cpu_nms: Enforce NMS to run on CPU.

  Returns:
    'nmsed_boxes': A [batch_size, max_detections, 4] float32 tensor
      containing the non-max suppressed boxes.
    'nmsed_scores': A [batch_size, max_detections] float32 tensor containing
      the scores for the boxes.
    'nmsed_classes': A [batch_size, max_detections] float32 tensor
      containing the class for boxes.
    'nmsed_masks': (optional) a
      [batch_size, max_detections, mask_height, mask_width] float32 tensor
      containing masks for each selected box. This is set to None if input
      `masks` is None.
    'nmsed_additional_fields': (optional) a dictionary of
      [batch_size, max_detections, ...] float32 tensors corresponding to the
      tensors specified in the input `additional_fields`. This is not returned
      if input `additional_fields` is None.
    'num_detections': A [batch_size] int32 tensor indicating the number of
      valid detections per batch item. Only the top num_detections[i] entries in
      nms_boxes[i], nms_scores[i] and nms_class[i] are valid. The rest of the
      entries are zero paddings.

  Raises:
    ValueError: if `q` in boxes.shape is not 1 or not equal to number of
      classes as inferred from scores.shape.
  """
  if use_combined_nms:
    if change_coordinate_frame:
      raise ValueError(
          'change_coordinate_frame (normalizing coordinates'
          ' relative to clip_window) is not supported by combined_nms.')
    if num_valid_boxes is not None:
      raise ValueError('num_valid_boxes is not supported by combined_nms.')
    if masks is not None:
      raise ValueError('masks is not supported by combined_nms.')
    if soft_nms_sigma != 0.0:
      raise ValueError('Soft NMS is not supported by combined_nms.')
    if use_class_agnostic_nms:
      raise ValueError('class-agnostic NMS is not supported by combined_nms.')
    if clip_window is not None:
      tf.logging.warning(
          'clip_window is not supported by combined_nms unless it is'
          ' [0. 0. 1. 1.] for each image.')
    if additional_fields is not None:
      tf.logging.warning('additional_fields is not supported by combined_nms.')
    if parallel_iterations != 32:
      tf.logging.warning('Number of batch items to be processed in parallel is'
                         ' not configurable by combined_nms.')
    if max_classes_per_detection > 1:
      tf.logging.warning(
          'max_classes_per_detection is not configurable by combined_nms.')

    with tf.name_scope(scope, 'CombinedNonMaxSuppression'):
      (batch_nmsed_boxes, batch_nmsed_scores, batch_nmsed_classes,
       batch_num_detections) = tf.image.combined_non_max_suppression(
           boxes=boxes,
           scores=scores,
           max_output_size_per_class=max_size_per_class,
           max_total_size=max_total_size,
           iou_threshold=iou_thresh,
           score_threshold=score_thresh,
           pad_per_class=use_static_shapes)
      # Not supported by combined_non_max_suppression.
      batch_nmsed_masks = None
      # Not supported by combined_non_max_suppression.
      batch_nmsed_additional_fields = None
      return (batch_nmsed_boxes, batch_nmsed_scores, batch_nmsed_classes,
              batch_nmsed_masks, batch_nmsed_additional_fields,
              batch_num_detections)

  q = shape_utils.get_dim_as_int(boxes.shape[2])
  num_classes = shape_utils.get_dim_as_int(scores.shape[2])
  if q != 1 and q != num_classes:
    raise ValueError('third dimension of boxes must be either 1 or equal '
                     'to the third dimension of scores.')
  if change_coordinate_frame and clip_window is None:
    raise ValueError('if change_coordinate_frame is True, then a clip_window'
                     'must be specified.')
  original_masks = masks

  # Create ordered dictionary using the sorted keys from
  # additional fields to ensure getting the same key value assignment
  # in _single_image_nms_fn(). The dictionary is thus a sorted version of
  # additional_fields.
  if additional_fields is None:
    ordered_additional_fields = collections.OrderedDict()
  else:
    ordered_additional_fields = collections.OrderedDict(
        sorted(additional_fields.items(), key=lambda item: item[0]))

  with tf.name_scope(scope, 'BatchMultiClassNonMaxSuppression'):
    boxes_shape = boxes.shape
    batch_size = shape_utils.get_dim_as_int(boxes_shape[0])
    num_anchors = shape_utils.get_dim_as_int(boxes_shape[1])

    if batch_size is None:
      batch_size = tf.shape(boxes)[0]
    if num_anchors is None:
      num_anchors = tf.shape(boxes)[1]

    # If num valid boxes aren't provided, create one and mark all boxes as
    # valid.
    if num_valid_boxes is None:
      num_valid_boxes = tf.ones([batch_size], dtype=tf.int32) * num_anchors

    # If masks aren't provided, create dummy masks so we can only have one copy
    # of _single_image_nms_fn and discard the dummy masks after map_fn.
    if masks is None:
      masks_shape = tf.stack([batch_size, num_anchors, q, 1, 1])
      masks = tf.zeros(masks_shape)

    if clip_window is None:
      clip_window = tf.stack([
          tf.reduce_min(boxes[:, :, :, 0]),
          tf.reduce_min(boxes[:, :, :, 1]),
          tf.reduce_max(boxes[:, :, :, 2]),
          tf.reduce_max(boxes[:, :, :, 3])
      ])
    if clip_window.shape.ndims == 1:
      clip_window = tf.tile(tf.expand_dims(clip_window, 0), [batch_size, 1])

    def _single_image_nms_fn(args):
      """Runs NMS on a single image and returns padded output.

      Args:
        args: A list of tensors consisting of the following:
          per_image_boxes - A [num_anchors, q, 4] float32 tensor containing
            detections. If `q` is 1 then same boxes are used for all classes
            otherwise, if `q` is equal to number of classes, class-specific
            boxes are used.
          per_image_scores - A [num_anchors, num_classes] float32 tensor
            containing the scores for each of the `num_anchors` detections.
          per_image_masks - A [num_anchors, q, mask_height, mask_width] float32
            tensor containing box masks. `q` can be either number of classes
            or 1 depending on whether a separate mask is predicted per class.
          per_image_clip_window - A 1D float32 tensor of the form
            [ymin, xmin, ymax, xmax] representing the window to clip the boxes
            to.
          per_image_additional_fields - (optional) A variable number of float32
            tensors each with size [num_anchors, ...].
          per_image_num_valid_boxes - A tensor of type `int32`. A 1-D tensor of
            shape [batch_size] representing the number of valid boxes to be
            considered for each image in the batch.  This parameter allows for
            ignoring zero paddings.

      Returns:
        'nmsed_boxes': A [max_detections, 4] float32 tensor containing the
          non-max suppressed boxes.
        'nmsed_scores': A [max_detections] float32 tensor containing the scores
          for the boxes.
        'nmsed_classes': A [max_detections] float32 tensor containing the class
          for boxes.
        'nmsed_masks': (optional) a [max_detections, mask_height, mask_width]
          float32 tensor containing masks for each selected box. This is set to
          None if input `masks` is None.
        'nmsed_additional_fields':  (optional) A variable number of float32
          tensors each with size [max_detections, ...] corresponding to the
          input `per_image_additional_fields`.
        'num_detections': A [batch_size] int32 tensor indicating the number of
          valid detections per batch item. Only the top num_detections[i]
          entries in nms_boxes[i], nms_scores[i] and nms_class[i] are valid. The
          rest of the entries are zero paddings.
      """
      per_image_boxes = args[0]
      per_image_scores = args[1]
      per_image_masks = args[2]
      per_image_clip_window = args[3]
      # Make sure that the order of elements passed in args is aligned with
      # the iteration order of ordered_additional_fields
      per_image_additional_fields = {
          key: value
          for key, value in zip(ordered_additional_fields, args[4:-1])
      }
      per_image_num_valid_boxes = args[-1]
      if use_static_shapes:
        total_proposals = tf.shape(per_image_scores)
        per_image_scores = tf.where(
            tf.less(tf.range(total_proposals[0]), per_image_num_valid_boxes),
            per_image_scores,
            tf.fill(total_proposals, np.finfo('float32').min))
      else:
        per_image_boxes = tf.reshape(
            tf.slice(per_image_boxes, 3 * [0],
                     tf.stack([per_image_num_valid_boxes, -1, -1])), [-1, q, 4])
        per_image_scores = tf.reshape(
            tf.slice(per_image_scores, [0, 0],
                     tf.stack([per_image_num_valid_boxes, -1])),
            [-1, num_classes])
        per_image_masks = tf.reshape(
            tf.slice(per_image_masks, 4 * [0],
                     tf.stack([per_image_num_valid_boxes, -1, -1, -1])),
            [-1, q, shape_utils.get_dim_as_int(per_image_masks.shape[2]),
             shape_utils.get_dim_as_int(per_image_masks.shape[3])])
        if per_image_additional_fields is not None:
          for key, tensor in per_image_additional_fields.items():
            additional_field_shape = tensor.get_shape()
            additional_field_dim = len(additional_field_shape)
            per_image_additional_fields[key] = tf.reshape(
                tf.slice(
                    per_image_additional_fields[key],
                    additional_field_dim * [0],
                    tf.stack([per_image_num_valid_boxes] +
                             (additional_field_dim - 1) * [-1])), [-1] + [
                                 shape_utils.get_dim_as_int(dim)
                                 for dim in additional_field_shape[1:]
                             ])
      if use_class_agnostic_nms:
        nmsed_boxlist, num_valid_nms_boxes = class_agnostic_non_max_suppression(
            per_image_boxes,
            per_image_scores,
            score_thresh,
            iou_thresh,
            max_classes_per_detection,
            max_total_size,
            clip_window=per_image_clip_window,
            change_coordinate_frame=change_coordinate_frame,
            masks=per_image_masks,
            pad_to_max_output_size=use_static_shapes,
            use_partitioned_nms=use_partitioned_nms,
            additional_fields=per_image_additional_fields,
            soft_nms_sigma=soft_nms_sigma)
      else:
        nmsed_boxlist, num_valid_nms_boxes = multiclass_non_max_suppression(
            per_image_boxes,
            per_image_scores,
            score_thresh,
            iou_thresh,
            max_size_per_class,
            max_total_size,
            clip_window=per_image_clip_window,
            change_coordinate_frame=change_coordinate_frame,
            masks=per_image_masks,
            pad_to_max_output_size=use_static_shapes,
            use_partitioned_nms=use_partitioned_nms,
            additional_fields=per_image_additional_fields,
            soft_nms_sigma=soft_nms_sigma,
            use_hard_nms=use_hard_nms,
            use_cpu_nms=use_cpu_nms)

      if not use_static_shapes:
        nmsed_boxlist = box_list_ops.pad_or_clip_box_list(
            nmsed_boxlist, max_total_size)
      num_detections = num_valid_nms_boxes
      nmsed_boxes = nmsed_boxlist.get()
      nmsed_scores = nmsed_boxlist.get_field(fields.BoxListFields.scores)
      nmsed_classes = nmsed_boxlist.get_field(fields.BoxListFields.classes)
      nmsed_masks = nmsed_boxlist.get_field(fields.BoxListFields.masks)
      nmsed_additional_fields = []
      # Sorting is needed here to ensure that the values stored in
      # nmsed_additional_fields are always kept in the same order
      # across different execution runs.
      for key in sorted(per_image_additional_fields.keys()):
        nmsed_additional_fields.append(nmsed_boxlist.get_field(key))
      return ([nmsed_boxes, nmsed_scores, nmsed_classes, nmsed_masks] +
              nmsed_additional_fields + [num_detections])

    num_additional_fields = 0
    if ordered_additional_fields:
      num_additional_fields = len(ordered_additional_fields)
    num_nmsed_outputs = 4 + num_additional_fields

    if use_dynamic_map_fn:
      map_fn = tf.map_fn
    else:
      map_fn = shape_utils.static_or_dynamic_map_fn

    batch_outputs = map_fn(
        _single_image_nms_fn,
        elems=([boxes, scores, masks, clip_window] +
               list(ordered_additional_fields.values()) + [num_valid_boxes]),
        dtype=(num_nmsed_outputs * [tf.float32] + [tf.int32]),
        parallel_iterations=parallel_iterations)

    batch_nmsed_boxes = batch_outputs[0]
    batch_nmsed_scores = batch_outputs[1]
    batch_nmsed_classes = batch_outputs[2]
    batch_nmsed_masks = batch_outputs[3]
    batch_nmsed_values = batch_outputs[4:-1]

    batch_nmsed_additional_fields = {}
    if num_additional_fields > 0:
      # Sort the keys to ensure arranging elements in same order as
      # in _single_image_nms_fn.
      batch_nmsed_keys = list(ordered_additional_fields.keys())
      for i in range(len(batch_nmsed_keys)):
        batch_nmsed_additional_fields[
            batch_nmsed_keys[i]] = batch_nmsed_values[i]

    batch_num_detections = batch_outputs[-1]

    if original_masks is None:
      batch_nmsed_masks = None

    if not ordered_additional_fields:
      batch_nmsed_additional_fields = None

    return (batch_nmsed_boxes, batch_nmsed_scores, batch_nmsed_classes,
            batch_nmsed_masks, batch_nmsed_additional_fields,
            batch_num_detections)



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Tests for google3.research.vale.object_detection.losses."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math

import numpy as np
from six.moves import zip
import tensorflow.compat.v1 as tf

from object_detection.core import box_list
from object_detection.core import losses
from object_detection.core import matcher
from object_detection.utils import test_case


class WeightedL2LocalizationLossTest(test_case.TestCase):

  def testReturnsCorrectWeightedLoss(self):
    batch_size = 3
    num_anchors = 10
    code_size = 4
    def graph_fn():
      prediction_tensor = tf.ones([batch_size, num_anchors, code_size])
      target_tensor = tf.zeros([batch_size, num_anchors, code_size])
      weights = tf.constant([[1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
                             [1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
                             [1, 1, 1, 1, 1, 0, 0, 0, 0, 0]], tf.float32)
      loss_op = losses.WeightedL2LocalizationLoss()
      loss = tf.reduce_sum(loss_op(prediction_tensor, target_tensor,
                                   weights=weights))
      return loss
    expected_loss = (3 * 5 * 4) / 2.0
    loss_output = self.execute(graph_fn, [])
    self.assertAllClose(loss_output, expected_loss)

  def testReturnsCorrectAnchorwiseLoss(self):
    batch_size = 3
    num_anchors = 16
    code_size = 4
    def graph_fn():
      prediction_tensor = tf.ones([batch_size, num_anchors, code_size])
      target_tensor = tf.zeros([batch_size, num_anchors, code_size])
      weights = tf.ones([batch_size, num_anchors])
      loss_op = losses.WeightedL2LocalizationLoss()
      loss = loss_op(prediction_tensor, target_tensor, weights=weights)
      return loss
    expected_loss = np.ones((batch_size, num_anchors)) * 2
    loss_output = self.execute(graph_fn, [])
    self.assertAllClose(loss_output, expected_loss)

  def testReturnsCorrectNanLoss(self):
    batch_size = 3
    num_anchors = 10
    code_size = 4
    def graph_fn():
      prediction_tensor = tf.ones([batch_size, num_anchors, code_size])
      target_tensor = tf.concat([
          tf.zeros([batch_size, num_anchors, code_size / 2]),
          tf.ones([batch_size, num_anchors, code_size / 2]) * np.nan
      ], axis=2)
      weights = tf.ones([batch_size, num_anchors])
      loss_op = losses.WeightedL2LocalizationLoss()
      loss = loss_op(prediction_tensor, target_tensor, weights=weights,
                     ignore_nan_targets=True)
      loss = tf.reduce_sum(loss)
      return loss
    expected_loss = (3 * 5 * 4) / 2.0
    loss_output = self.execute(graph_fn, [])
    self.assertAllClose(loss_output, expected_loss)

  def testReturnsCorrectWeightedLossWithLossesMask(self):
    batch_size = 4
    num_anchors = 10
    code_size = 4
    def graph_fn():
      prediction_tensor = tf.ones([batch_size, num_anchors, code_size])
      target_tensor = tf.zeros([batch_size, num_anchors, code_size])
      weights = tf.constant([[1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
                             [1, 1, 1, 1, 1, 1, 1, 1, 0, 0],
                             [1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
                             [1, 1, 1, 1, 1, 0, 0, 0, 0, 0]], tf.float32)
      losses_mask = tf.constant([True, False, True, True], tf.bool)
      loss_op = losses.WeightedL2LocalizationLoss()
      loss = tf.reduce_sum(loss_op(prediction_tensor, target_tensor,
                                   weights=weights, losses_mask=losses_mask))
      return loss
    expected_loss = (3 * 5 * 4) / 2.0
    loss_output = self.execute(graph_fn, [])
    self.assertAllClose(loss_output, expected_loss)


class WeightedSmoothL1LocalizationLossTest(test_case.TestCase):

  def testReturnsCorrectLoss(self):
    batch_size = 2
    num_anchors = 3
    code_size = 4
    def graph_fn():
      prediction_tensor = tf.constant([[[2.5, 0, .4, 0],
                                        [0, 0, 0, 0],
                                        [0, 2.5, 0, .4]],
                                       [[3.5, 0, 0, 0],
                                        [0, .4, 0, .9],
                                        [0, 0, 1.5, 0]]], tf.float32)
      target_tensor = tf.zeros([batch_size, num_anchors, code_size])
      weights = tf.constant([[2, 1, 1],
                             [0, 3, 0]], tf.float32)
      loss_op = losses.WeightedSmoothL1LocalizationLoss()
      loss = loss_op(prediction_tensor, target_tensor, weights=weights)
      loss = tf.reduce_sum(loss)
      return loss
    exp_loss = 7.695
    loss_output = self.execute(graph_fn, [])
    self.assertAllClose(loss_output, exp_loss)

  def testReturnsCorrectLossWithLossesMask(self):
    batch_size = 3
    num_anchors = 3
    code_size = 4
    def graph_fn():
      prediction_tensor = tf.constant([[[2.5, 0, .4, 0],
                                        [0, 0, 0, 0],
                                        [0, 2.5, 0, .4]],
                                       [[3.5, 0, 0, 0],
                                        [0, .4, 0, .9],
                                        [0, 0, 1.5, 0]],
                                       [[3.5, 7., 0, 0],
                                        [0, .4, 0, .9],
                                        [2.2, 2.2, 1.5, 0]]], tf.float32)
      target_tensor = tf.zeros([batch_size, num_anchors, code_size])
      weights = tf.constant([[2, 1, 1],
                             [0, 3, 0],
                             [4, 3, 0]], tf.float32)
      losses_mask = tf.constant([True, True, False], tf.bool)
      loss_op = losses.WeightedSmoothL1LocalizationLoss()
      loss = loss_op(prediction_tensor, target_tensor, weights=weights,
                     losses_mask=losses_mask)
      loss = tf.reduce_sum(loss)
      return loss
    exp_loss = 7.695
    loss_output = self.execute(graph_fn, [])
    self.assertAllClose(loss_output, exp_loss)


class WeightedIOULocalizationLossTest(test_case.TestCase):

  def testReturnsCorrectLoss(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[1.5, 0, 2.4, 1],
                                        [0, 0, 1, 1],
                                        [0, 0, .5, .25]]])
      target_tensor = tf.constant([[[1.5, 0, 2.4, 1],
                                    [0, 0, 1, 1],
                                    [50, 50, 500.5, 100.25]]])
      weights = [[1.0, .5, 2.0]]
      loss_op = losses.WeightedIOULocalizationLoss()
      loss = loss_op(prediction_tensor, target_tensor, weights=weights)
      loss = tf.reduce_sum(loss)
      return loss
    exp_loss = 2.0
    loss_output = self.execute(graph_fn, [])
    self.assertAllClose(loss_output, exp_loss)

  def testReturnsCorrectLossWithNoLabels(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[1.5, 0, 2.4, 1],
                                        [0, 0, 1, 1],
                                        [0, 0, .5, .25]]])
      target_tensor = tf.constant([[[1.5, 0, 2.4, 1],
                                    [0, 0, 1, 1],
                                    [50, 50, 500.5, 100.25]]])
      weights = [[1.0, .5, 2.0]]
      losses_mask = tf.constant([False], tf.bool)
      loss_op = losses.WeightedIOULocalizationLoss()
      loss = loss_op(prediction_tensor, target_tensor, weights=weights,
                     losses_mask=losses_mask)
      loss = tf.reduce_sum(loss)
      return loss
    exp_loss = 0.0
    loss_output = self.execute(graph_fn, [])
    self.assertAllClose(loss_output, exp_loss)


class WeightedGIOULocalizationLossTest(test_case.TestCase):

  def testReturnsCorrectLoss(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[1.5, 0, 2.4, 1],
                                        [0, 0, 1, 1],
                                        [0, 0, 0, 0]]])
      target_tensor = tf.constant([[[1.5, 0, 2.4, 1],
                                    [0, 0, 1, 1],
                                    [5, 5, 10, 10]]])
      weights = [[1.0, .5, 2.0]]
      loss_op = losses.WeightedGIOULocalizationLoss()
      loss = loss_op(prediction_tensor,
                     target_tensor,
                     weights=weights)
      loss = tf.reduce_sum(loss)
      return loss
    exp_loss = 3.5
    loss_output = self.execute(graph_fn, [])
    self.assertAllClose(loss_output, exp_loss)

  def testReturnsCorrectLossWithNoLabels(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[1.5, 0, 2.4, 1],
                                        [0, 0, 1, 1],
                                        [0, 0, .5, .25]]])
      target_tensor = tf.constant([[[1.5, 0, 2.4, 1],
                                    [0, 0, 1, 1],
                                    [50, 50, 500.5, 100.25]]])
      weights = [[1.0, .5, 2.0]]
      losses_mask = tf.constant([False], tf.bool)
      loss_op = losses.WeightedGIOULocalizationLoss()
      loss = loss_op(prediction_tensor, target_tensor, weights=weights,
                     losses_mask=losses_mask)
      loss = tf.reduce_sum(loss)
      return loss
    exp_loss = 0.0
    loss_output = self.execute(graph_fn, [])
    self.assertAllClose(loss_output, exp_loss)


class WeightedSigmoidClassificationLossTest(test_case.TestCase):

  def testReturnsCorrectLoss(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[-100, 100, -100],
                                        [100, -100, -100],
                                        [100, 0, -100],
                                        [-100, -100, 100]],
                                       [[-100, 0, 100],
                                        [-100, 100, -100],
                                        [100, 100, 100],
                                        [0, 0, -1]]], tf.float32)
      target_tensor = tf.constant([[[0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0],
                                    [0, 0, 1]],
                                   [[0, 0, 1],
                                    [0, 1, 0],
                                    [1, 1, 1],
                                    [1, 0, 0]]], tf.float32)
      weights = tf.constant([[[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [0, 0, 0]]], tf.float32)
      loss_op = losses.WeightedSigmoidClassificationLoss()
      loss = loss_op(prediction_tensor, target_tensor, weights=weights)
      loss = tf.reduce_sum(loss)
      return loss

    exp_loss = -2 * math.log(.5)
    loss_output = self.execute(graph_fn, [])
    self.assertAllClose(loss_output, exp_loss)

  def testReturnsCorrectAnchorWiseLoss(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[-100, 100, -100],
                                        [100, -100, -100],
                                        [100, 0, -100],
                                        [-100, -100, 100]],
                                       [[-100, 0, 100],
                                        [-100, 100, -100],
                                        [100, 100, 100],
                                        [0, 0, -1]]], tf.float32)
      target_tensor = tf.constant([[[0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0],
                                    [0, 0, 1]],
                                   [[0, 0, 1],
                                    [0, 1, 0],
                                    [1, 1, 1],
                                    [1, 0, 0]]], tf.float32)
      weights = tf.constant([[[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [0, 0, 0]]], tf.float32)
      loss_op = losses.WeightedSigmoidClassificationLoss()
      loss = loss_op(prediction_tensor, target_tensor, weights=weights)
      loss = tf.reduce_sum(loss, axis=2)
      return loss

    exp_loss = np.matrix([[0, 0, -math.log(.5), 0],
                          [-math.log(.5), 0, 0, 0]])
    loss_output = self.execute(graph_fn, [])
    self.assertAllClose(loss_output, exp_loss)

  def testReturnsCorrectLossWithClassIndices(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[-100, 100, -100, 100],
                                        [100, -100, -100, -100],
                                        [100, 0, -100, 100],
                                        [-100, -100, 100, -100]],
                                       [[-100, 0, 100, 100],
                                        [-100, 100, -100, 100],
                                        [100, 100, 100, 100],
                                        [0, 0, -1, 100]]], tf.float32)
      target_tensor = tf.constant([[[0, 1, 0, 0],
                                    [1, 0, 0, 1],
                                    [1, 0, 0, 0],
                                    [0, 0, 1, 1]],
                                   [[0, 0, 1, 0],
                                    [0, 1, 0, 0],
                                    [1, 1, 1, 0],
                                    [1, 0, 0, 0]]], tf.float32)
      weights = tf.constant([[[1, 1, 1, 1],
                              [1, 1, 1, 1],
                              [1, 1, 1, 1],
                              [1, 1, 1, 1]],
                             [[1, 1, 1, 1],
                              [1, 1, 1, 1],
                              [1, 1, 1, 1],
                              [0, 0, 0, 0]]], tf.float32)
      # Ignores the last class.
      class_indices = tf.constant([0, 1, 2], tf.int32)
      loss_op = losses.WeightedSigmoidClassificationLoss()
      loss = loss_op(prediction_tensor, target_tensor, weights=weights,
                     class_indices=class_indices)
      loss = tf.reduce_sum(loss, axis=2)
      return loss

    exp_loss = np.matrix([[0, 0, -math.log(.5), 0],
                          [-math.log(.5), 0, 0, 0]])
    loss_output = self.execute(graph_fn, [])
    self.assertAllClose(loss_output, exp_loss)

  def testReturnsCorrectLossWithLossesMask(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[-100, 100, -100],
                                        [100, -100, -100],
                                        [100, 0, -100],
                                        [-100, -100, 100]],
                                       [[-100, 0, 100],
                                        [-100, 100, -100],
                                        [100, 100, 100],
                                        [0, 0, -1]],
                                       [[-100, 0, 100],
                                        [-100, 100, -100],
                                        [100, 100, 100],
                                        [0, 0, -100]]], tf.float32)
      target_tensor = tf.constant([[[0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0],
                                    [0, 0, 1]],
                                   [[0, 0, 1],
                                    [0, 1, 0],
                                    [1, 1, 1],
                                    [1, 0, 0]],
                                   [[0, 0, 0],
                                    [0, 0, 0],
                                    [0, 0, 0],
                                    [0, 0, 0]]], tf.float32)
      weights = tf.constant([[[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [0, 0, 0]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]]], tf.float32)
      losses_mask = tf.constant([True, True, False], tf.bool)

      loss_op = losses.WeightedSigmoidClassificationLoss()
      loss_per_anchor = loss_op(prediction_tensor, target_tensor,
                                weights=weights,
                                losses_mask=losses_mask)
      loss = tf.reduce_sum(loss_per_anchor)
      return loss

    exp_loss = -2 * math.log(.5)
    loss_output = self.execute(graph_fn, [])
    self.assertAllClose(loss_output, exp_loss)


def _logit(probability):
  return math.log(probability / (1. - probability))


class SigmoidFocalClassificationLossTest(test_case.TestCase):

  def testEasyExamplesProduceSmallLossComparedToSigmoidXEntropy(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[_logit(0.97)],
                                        [_logit(0.91)],
                                        [_logit(0.73)],
                                        [_logit(0.27)],
                                        [_logit(0.09)],
                                        [_logit(0.03)]]], tf.float32)
      target_tensor = tf.constant([[[1],
                                    [1],
                                    [1],
                                    [0],
                                    [0],
                                    [0]]], tf.float32)
      weights = tf.constant([[[1], [1], [1], [1], [1], [1]]], tf.float32)
      focal_loss_op = losses.SigmoidFocalClassificationLoss(gamma=2.0,
                                                            alpha=None)
      sigmoid_loss_op = losses.WeightedSigmoidClassificationLoss()
      focal_loss = tf.reduce_sum(focal_loss_op(prediction_tensor, target_tensor,
                                               weights=weights), axis=2)
      sigmoid_loss = tf.reduce_sum(sigmoid_loss_op(prediction_tensor,
                                                   target_tensor,
                                                   weights=weights), axis=2)
      return sigmoid_loss, focal_loss

    sigmoid_loss, focal_loss = self.execute(graph_fn, [])
    order_of_ratio = np.power(10,
                              np.floor(np.log10(sigmoid_loss / focal_loss)))
    self.assertAllClose(order_of_ratio, [[1000, 100, 10, 10, 100, 1000]])

  def testHardExamplesProduceLossComparableToSigmoidXEntropy(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[_logit(0.55)],
                                        [_logit(0.52)],
                                        [_logit(0.50)],
                                        [_logit(0.48)],
                                        [_logit(0.45)]]], tf.float32)
      target_tensor = tf.constant([[[1],
                                    [1],
                                    [1],
                                    [0],
                                    [0]]], tf.float32)
      weights = tf.constant([[[1], [1], [1], [1], [1]]], tf.float32)
      focal_loss_op = losses.SigmoidFocalClassificationLoss(gamma=2.0,
                                                            alpha=None)
      sigmoid_loss_op = losses.WeightedSigmoidClassificationLoss()
      focal_loss = tf.reduce_sum(focal_loss_op(prediction_tensor, target_tensor,
                                               weights=weights), axis=2)
      sigmoid_loss = tf.reduce_sum(sigmoid_loss_op(prediction_tensor,
                                                   target_tensor,
                                                   weights=weights), axis=2)
      return sigmoid_loss, focal_loss
    sigmoid_loss, focal_loss = self.execute(graph_fn, [])
    order_of_ratio = np.power(10,
                              np.floor(np.log10(sigmoid_loss / focal_loss)))
    self.assertAllClose(order_of_ratio, [[1., 1., 1., 1., 1.]])

  def testNonAnchorWiseOutputComparableToSigmoidXEntropy(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[_logit(0.55)],
                                        [_logit(0.52)],
                                        [_logit(0.50)],
                                        [_logit(0.48)],
                                        [_logit(0.45)]]], tf.float32)
      target_tensor = tf.constant([[[1],
                                    [1],
                                    [1],
                                    [0],
                                    [0]]], tf.float32)
      weights = tf.constant([[[1], [1], [1], [1], [1]]], tf.float32)
      focal_loss_op = losses.SigmoidFocalClassificationLoss(gamma=2.0,
                                                            alpha=None)
      sigmoid_loss_op = losses.WeightedSigmoidClassificationLoss()
      focal_loss = tf.reduce_sum(focal_loss_op(prediction_tensor, target_tensor,
                                               weights=weights))
      sigmoid_loss = tf.reduce_sum(sigmoid_loss_op(prediction_tensor,
                                                   target_tensor,
                                                   weights=weights))
      return sigmoid_loss, focal_loss
    sigmoid_loss, focal_loss = self.execute(graph_fn, [])
    order_of_ratio = np.power(10,
                              np.floor(np.log10(sigmoid_loss / focal_loss)))
    self.assertAlmostEqual(order_of_ratio, 1.)

  def testIgnoreNegativeExampleLossViaAlphaMultiplier(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[_logit(0.55)],
                                        [_logit(0.52)],
                                        [_logit(0.50)],
                                        [_logit(0.48)],
                                        [_logit(0.45)]]], tf.float32)
      target_tensor = tf.constant([[[1],
                                    [1],
                                    [1],
                                    [0],
                                    [0]]], tf.float32)
      weights = tf.constant([[[1], [1], [1], [1], [1]]], tf.float32)
      focal_loss_op = losses.SigmoidFocalClassificationLoss(gamma=2.0,
                                                            alpha=1.0)
      sigmoid_loss_op = losses.WeightedSigmoidClassificationLoss()
      focal_loss = tf.reduce_sum(focal_loss_op(prediction_tensor, target_tensor,
                                               weights=weights), axis=2)
      sigmoid_loss = tf.reduce_sum(sigmoid_loss_op(prediction_tensor,
                                                   target_tensor,
                                                   weights=weights), axis=2)
      return sigmoid_loss, focal_loss

    sigmoid_loss, focal_loss = self.execute(graph_fn, [])
    self.assertAllClose(focal_loss[0][3:], [0., 0.])
    order_of_ratio = np.power(10,
                              np.floor(np.log10(sigmoid_loss[0][:3] /
                                                focal_loss[0][:3])))
    self.assertAllClose(order_of_ratio, [1., 1., 1.])

  def testIgnorePositiveExampleLossViaAlphaMultiplier(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[_logit(0.55)],
                                        [_logit(0.52)],
                                        [_logit(0.50)],
                                        [_logit(0.48)],
                                        [_logit(0.45)]]], tf.float32)
      target_tensor = tf.constant([[[1],
                                    [1],
                                    [1],
                                    [0],
                                    [0]]], tf.float32)
      weights = tf.constant([[[1], [1], [1], [1], [1]]], tf.float32)
      focal_loss_op = losses.SigmoidFocalClassificationLoss(gamma=2.0,
                                                            alpha=0.0)
      sigmoid_loss_op = losses.WeightedSigmoidClassificationLoss()
      focal_loss = tf.reduce_sum(focal_loss_op(prediction_tensor, target_tensor,
                                               weights=weights), axis=2)
      sigmoid_loss = tf.reduce_sum(sigmoid_loss_op(prediction_tensor,
                                                   target_tensor,
                                                   weights=weights), axis=2)
      return sigmoid_loss, focal_loss
    sigmoid_loss, focal_loss = self.execute(graph_fn, [])
    self.assertAllClose(focal_loss[0][:3], [0., 0., 0.])
    order_of_ratio = np.power(10,
                              np.floor(np.log10(sigmoid_loss[0][3:] /
                                                focal_loss[0][3:])))
    self.assertAllClose(order_of_ratio, [1., 1.])

  def testSimilarToSigmoidXEntropyWithHalfAlphaAndZeroGammaUpToAScale(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[-100, 100, -100],
                                        [100, -100, -100],
                                        [100, 0, -100],
                                        [-100, -100, 100]],
                                       [[-100, 0, 100],
                                        [-100, 100, -100],
                                        [100, 100, 100],
                                        [0, 0, -1]]], tf.float32)
      target_tensor = tf.constant([[[0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0],
                                    [0, 0, 1]],
                                   [[0, 0, 1],
                                    [0, 1, 0],
                                    [1, 1, 1],
                                    [1, 0, 0]]], tf.float32)
      weights = tf.constant([[[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [0, 0, 0]]], tf.float32)
      focal_loss_op = losses.SigmoidFocalClassificationLoss(alpha=0.5,
                                                            gamma=0.0)
      sigmoid_loss_op = losses.WeightedSigmoidClassificationLoss()
      focal_loss = focal_loss_op(prediction_tensor, target_tensor,
                                 weights=weights)
      sigmoid_loss = sigmoid_loss_op(prediction_tensor, target_tensor,
                                     weights=weights)
      return sigmoid_loss, focal_loss
    sigmoid_loss, focal_loss = self.execute(graph_fn, [])
    self.assertAllClose(sigmoid_loss, focal_loss * 2)

  def testSameAsSigmoidXEntropyWithNoAlphaAndZeroGamma(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[-100, 100, -100],
                                        [100, -100, -100],
                                        [100, 0, -100],
                                        [-100, -100, 100]],
                                       [[-100, 0, 100],
                                        [-100, 100, -100],
                                        [100, 100, 100],
                                        [0, 0, -1]]], tf.float32)
      target_tensor = tf.constant([[[0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0],
                                    [0, 0, 1]],
                                   [[0, 0, 1],
                                    [0, 1, 0],
                                    [1, 1, 1],
                                    [1, 0, 0]]], tf.float32)
      weights = tf.constant([[[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [0, 0, 0]]], tf.float32)
      focal_loss_op = losses.SigmoidFocalClassificationLoss(alpha=None,
                                                            gamma=0.0)
      sigmoid_loss_op = losses.WeightedSigmoidClassificationLoss()
      focal_loss = focal_loss_op(prediction_tensor, target_tensor,
                                 weights=weights)
      sigmoid_loss = sigmoid_loss_op(prediction_tensor, target_tensor,
                                     weights=weights)
      return sigmoid_loss, focal_loss
    sigmoid_loss, focal_loss = self.execute(graph_fn, [])
    self.assertAllClose(sigmoid_loss, focal_loss)

  def testExpectedLossWithAlphaOneAndZeroGamma(self):
    def graph_fn():
      # All zeros correspond to 0.5 probability.
      prediction_tensor = tf.constant([[[0, 0, 0],
                                        [0, 0, 0],
                                        [0, 0, 0],
                                        [0, 0, 0]],
                                       [[0, 0, 0],
                                        [0, 0, 0],
                                        [0, 0, 0],
                                        [0, 0, 0]]], tf.float32)
      target_tensor = tf.constant([[[0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0],
                                    [0, 0, 1]],
                                   [[0, 0, 1],
                                    [0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0]]], tf.float32)
      weights = tf.constant([[[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]]], tf.float32)
      focal_loss_op = losses.SigmoidFocalClassificationLoss(alpha=1.0,
                                                            gamma=0.0)

      focal_loss = tf.reduce_sum(focal_loss_op(prediction_tensor, target_tensor,
                                               weights=weights))
      return focal_loss
    focal_loss = self.execute(graph_fn, [])
    self.assertAllClose(
        (-math.log(.5) *  # x-entropy per class per anchor
         1.0 *            # alpha
         8),              # positives from 8 anchors
        focal_loss)

  def testExpectedLossWithAlpha75AndZeroGamma(self):
    def graph_fn():
      # All zeros correspond to 0.5 probability.
      prediction_tensor = tf.constant([[[0, 0, 0],
                                        [0, 0, 0],
                                        [0, 0, 0],
                                        [0, 0, 0]],
                                       [[0, 0, 0],
                                        [0, 0, 0],
                                        [0, 0, 0],
                                        [0, 0, 0]]], tf.float32)
      target_tensor = tf.constant([[[0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0],
                                    [0, 0, 1]],
                                   [[0, 0, 1],
                                    [0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0]]], tf.float32)
      weights = tf.constant([[[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]]], tf.float32)
      focal_loss_op = losses.SigmoidFocalClassificationLoss(alpha=0.75,
                                                            gamma=0.0)

      focal_loss = tf.reduce_sum(focal_loss_op(prediction_tensor, target_tensor,
                                               weights=weights))
      return focal_loss
    focal_loss = self.execute(graph_fn, [])
    self.assertAllClose(
        (-math.log(.5) *  # x-entropy per class per anchor.
         ((0.75 *         # alpha for positives.
           8) +           # positives from 8 anchors.
          (0.25 *         # alpha for negatives.
           8 * 2))),      # negatives from 8 anchors for two classes.
        focal_loss)

  def testExpectedLossWithLossesMask(self):
    def graph_fn():
      # All zeros correspond to 0.5 probability.
      prediction_tensor = tf.constant([[[0, 0, 0],
                                        [0, 0, 0],
                                        [0, 0, 0],
                                        [0, 0, 0]],
                                       [[0, 0, 0],
                                        [0, 0, 0],
                                        [0, 0, 0],
                                        [0, 0, 0]],
                                       [[0, 0, 0],
                                        [0, 0, 0],
                                        [0, 0, 0],
                                        [0, 0, 0]]], tf.float32)
      target_tensor = tf.constant([[[0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0],
                                    [0, 0, 1]],
                                   [[0, 0, 1],
                                    [0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0]],
                                   [[1, 0, 0],
                                    [1, 0, 0],
                                    [1, 0, 0],
                                    [1, 0, 0]]], tf.float32)
      weights = tf.constant([[[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]]], tf.float32)
      losses_mask = tf.constant([True, True, False], tf.bool)
      focal_loss_op = losses.SigmoidFocalClassificationLoss(alpha=0.75,
                                                            gamma=0.0)

      focal_loss = tf.reduce_sum(focal_loss_op(prediction_tensor, target_tensor,
                                               weights=weights,
                                               losses_mask=losses_mask))
      return focal_loss
    focal_loss = self.execute(graph_fn, [])
    self.assertAllClose(
        (-math.log(.5) *  # x-entropy per class per anchor.
         ((0.75 *         # alpha for positives.
           8) +           # positives from 8 anchors.
          (0.25 *         # alpha for negatives.
           8 * 2))),      # negatives from 8 anchors for two classes.
        focal_loss)


class WeightedSoftmaxClassificationLossTest(test_case.TestCase):

  def testReturnsCorrectLoss(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[-100, 100, -100],
                                        [100, -100, -100],
                                        [0, 0, -100],
                                        [-100, -100, 100]],
                                       [[-100, 0, 0],
                                        [-100, 100, -100],
                                        [-100, 100, -100],
                                        [100, -100, -100]]], tf.float32)
      target_tensor = tf.constant([[[0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0],
                                    [0, 0, 1]],
                                   [[0, 0, 1],
                                    [0, 1, 0],
                                    [0, 1, 0],
                                    [1, 0, 0]]], tf.float32)
      weights = tf.constant([[[1, 1, 1],
                              [1, 1, 1],
                              [0.5, 0.5, 0.5],
                              [1, 1, 1]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [0, 0, 0]]], tf.float32)
      loss_op = losses.WeightedSoftmaxClassificationLoss()
      loss = loss_op(prediction_tensor, target_tensor, weights=weights)
      loss = tf.reduce_sum(loss)
      return loss
    loss_output = self.execute(graph_fn, [])
    exp_loss = - 1.5 * math.log(.5)
    self.assertAllClose(loss_output, exp_loss)

  def testReturnsCorrectAnchorWiseLoss(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[-100, 100, -100],
                                        [100, -100, -100],
                                        [0, 0, -100],
                                        [-100, -100, 100]],
                                       [[-100, 0, 0],
                                        [-100, 100, -100],
                                        [-100, 100, -100],
                                        [100, -100, -100]]], tf.float32)
      target_tensor = tf.constant([[[0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0],
                                    [0, 0, 1]],
                                   [[0, 0, 1],
                                    [0, 1, 0],
                                    [0, 1, 0],
                                    [1, 0, 0]]], tf.float32)
      weights = tf.constant([[[1, 1, 1],
                              [1, 1, 1],
                              [0.5, 0.5, 0.5],
                              [1, 1, 1]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [0, 0, 0]]], tf.float32)
      loss_op = losses.WeightedSoftmaxClassificationLoss()
      loss = loss_op(prediction_tensor, target_tensor, weights=weights)
      return loss
    loss_output = self.execute(graph_fn, [])
    exp_loss = np.matrix([[0, 0, - 0.5 * math.log(.5), 0],
                          [-math.log(.5), 0, 0, 0]])
    self.assertAllClose(loss_output, exp_loss)

  def testReturnsCorrectAnchorWiseLossWithHighLogitScaleSetting(self):
    """At very high logit_scale, all predictions will be ~0.33."""
    def graph_fn():
      # TODO(yonib): Also test logit_scale with anchorwise=False.
      logit_scale = 10e16
      prediction_tensor = tf.constant([[[-100, 100, -100],
                                        [100, -100, -100],
                                        [0, 0, -100],
                                        [-100, -100, 100]],
                                       [[-100, 0, 0],
                                        [-100, 100, -100],
                                        [-100, 100, -100],
                                        [100, -100, -100]]], tf.float32)
      target_tensor = tf.constant([[[0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0],
                                    [0, 0, 1]],
                                   [[0, 0, 1],
                                    [0, 1, 0],
                                    [0, 1, 0],
                                    [1, 0, 0]]], tf.float32)
      weights = tf.constant([[[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]]], tf.float32)
      loss_op = losses.WeightedSoftmaxClassificationLoss(
          logit_scale=logit_scale)
      loss = loss_op(prediction_tensor, target_tensor, weights=weights)
      return loss
    uniform_distribution_loss = - math.log(.33333333333)
    exp_loss = np.matrix([[uniform_distribution_loss] * 4,
                          [uniform_distribution_loss] * 4])
    loss_output = self.execute(graph_fn, [])
    self.assertAllClose(loss_output, exp_loss)

  def testReturnsCorrectLossWithLossesMask(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[-100, 100, -100],
                                        [100, -100, -100],
                                        [0, 0, -100],
                                        [-100, -100, 100]],
                                       [[-100, 0, 0],
                                        [-100, 100, -100],
                                        [-100, 100, -100],
                                        [100, -100, -100]],
                                       [[-100, 0, 0],
                                        [-100, 100, -100],
                                        [-100, 100, -100],
                                        [100, -100, -100]]], tf.float32)
      target_tensor = tf.constant([[[0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0],
                                    [0, 0, 1]],
                                   [[0, 0, 1],
                                    [0, 1, 0],
                                    [0, 1, 0],
                                    [1, 0, 0]],
                                   [[1, 0, 0],
                                    [1, 0, 0],
                                    [1, 0, 0],
                                    [1, 0, 0]]], tf.float32)
      weights = tf.constant([[[1, 1, 1],
                              [1, 1, 1],
                              [0.5, 0.5, 0.5],
                              [1, 1, 1]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [0, 0, 0]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]]], tf.float32)
      losses_mask = tf.constant([True, True, False], tf.bool)
      loss_op = losses.WeightedSoftmaxClassificationLoss()
      loss = loss_op(prediction_tensor, target_tensor, weights=weights,
                     losses_mask=losses_mask)
      loss = tf.reduce_sum(loss)
      return loss
    loss_output = self.execute(graph_fn, [])
    exp_loss = - 1.5 * math.log(.5)
    self.assertAllClose(loss_output, exp_loss)


class WeightedSoftmaxClassificationAgainstLogitsLossTest(test_case.TestCase):

  def testReturnsCorrectLoss(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[-100, 100, -100],
                                        [100, -100, -100],
                                        [0, 0, -100],
                                        [-100, -100, 100]],
                                       [[-100, 0, 0],
                                        [-100, 100, -100],
                                        [-100, 100, -100],
                                        [100, -100, -100]]], tf.float32)

      target_tensor = tf.constant([[[-100, 100, -100],
                                    [100, -100, -100],
                                    [100, -100, -100],
                                    [-100, -100, 100]],
                                   [[-100, -100, 100],
                                    [-100, 100, -100],
                                    [-100, 100, -100],
                                    [100, -100, -100]]], tf.float32)
      weights = tf.constant([[1, 1, .5, 1],
                             [1, 1, 1, 1]], tf.float32)
      weights_shape = tf.shape(weights)
      weights_multiple = tf.concat(
          [tf.ones_like(weights_shape), tf.constant([3])],
          axis=0)
      weights = tf.tile(tf.expand_dims(weights, 2), weights_multiple)
      loss_op = losses.WeightedSoftmaxClassificationAgainstLogitsLoss()
      loss = loss_op(prediction_tensor, target_tensor, weights=weights)
      loss = tf.reduce_sum(loss)
      return loss
    loss_output = self.execute(graph_fn, [])
    exp_loss = - 1.5 * math.log(.5)
    self.assertAllClose(loss_output, exp_loss)

  def testReturnsCorrectAnchorWiseLoss(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[-100, 100, -100],
                                        [100, -100, -100],
                                        [0, 0, -100],
                                        [-100, -100, 100]],
                                       [[-100, 0, 0],
                                        [-100, 100, -100],
                                        [-100, 100, -100],
                                        [100, -100, -100]]], tf.float32)
      target_tensor = tf.constant([[[-100, 100, -100],
                                    [100, -100, -100],
                                    [100, -100, -100],
                                    [-100, -100, 100]],
                                   [[-100, -100, 100],
                                    [-100, 100, -100],
                                    [-100, 100, -100],
                                    [100, -100, -100]]], tf.float32)
      weights = tf.constant([[1, 1, .5, 1],
                             [1, 1, 1, 0]], tf.float32)
      weights_shape = tf.shape(weights)
      weights_multiple = tf.concat(
          [tf.ones_like(weights_shape), tf.constant([3])],
          axis=0)
      weights = tf.tile(tf.expand_dims(weights, 2), weights_multiple)
      loss_op = losses.WeightedSoftmaxClassificationAgainstLogitsLoss()
      loss = loss_op(prediction_tensor, target_tensor, weights=weights)
      return loss
    loss_output = self.execute(graph_fn, [])
    exp_loss = np.matrix([[0, 0, - 0.5 * math.log(.5), 0],
                          [-math.log(.5), 0, 0, 0]])
    self.assertAllClose(loss_output, exp_loss)

  def testReturnsCorrectAnchorWiseLossWithLogitScaleSetting(self):
    def graph_fn():
      logit_scale = 100.
      prediction_tensor = tf.constant([[[-100, 100, -100],
                                        [100, -100, -100],
                                        [0, 0, -100],
                                        [-100, -100, 100]],
                                       [[-100, 0, 0],
                                        [-100, 100, -100],
                                        [-100, 100, -100],
                                        [100, -100, -100]]], tf.float32)
      target_tensor = tf.constant([[[-100, 100, -100],
                                    [100, -100, -100],
                                    [0, 0, -100],
                                    [-100, -100, 100]],
                                   [[-100, 0, 0],
                                    [-100, 100, -100],
                                    [-100, 100, -100],
                                    [100, -100, -100]]], tf.float32)
      weights = tf.constant([[1, 1, .5, 1],
                             [1, 1, 1, 0]], tf.float32)
      weights_shape = tf.shape(weights)
      weights_multiple = tf.concat(
          [tf.ones_like(weights_shape), tf.constant([3])],
          axis=0)
      weights = tf.tile(tf.expand_dims(weights, 2), weights_multiple)
      loss_op = losses.WeightedSoftmaxClassificationAgainstLogitsLoss(
          logit_scale=logit_scale)
      loss = loss_op(prediction_tensor, target_tensor, weights=weights)
      return loss

    # find softmax of the two prediction types above
    softmax_pred1 = [np.exp(-1), np.exp(-1), np.exp(1)]
    softmax_pred1 /= sum(softmax_pred1)
    softmax_pred2 = [np.exp(0), np.exp(0), np.exp(-1)]
    softmax_pred2 /= sum(softmax_pred2)

    # compute the expected cross entropy for perfect matches
    exp_entropy1 = sum(
        [-x*np.log(x) for x in softmax_pred1])
    exp_entropy2 = sum(
        [-x*np.log(x) for x in softmax_pred2])

    # weighted expected losses
    exp_loss = np.matrix(
        [[exp_entropy1, exp_entropy1, exp_entropy2*.5, exp_entropy1],
         [exp_entropy2, exp_entropy1, exp_entropy1, 0.]])
    loss_output = self.execute(graph_fn, [])
    self.assertAllClose(loss_output, exp_loss)


class BootstrappedSigmoidClassificationLossTest(test_case.TestCase):

  def testReturnsCorrectLossSoftBootstrapping(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[-100, 100, 0],
                                        [100, -100, -100],
                                        [100, -100, -100],
                                        [-100, -100, 100]],
                                       [[-100, -100, 100],
                                        [-100, 100, -100],
                                        [100, 100, 100],
                                        [0, 0, -1]]], tf.float32)
      target_tensor = tf.constant([[[0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0],
                                    [0, 0, 1]],
                                   [[0, 0, 1],
                                    [0, 1, 0],
                                    [1, 1, 1],
                                    [1, 0, 0]]], tf.float32)
      weights = tf.constant([[[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [0, 0, 0]]], tf.float32)
      alpha = tf.constant(.5, tf.float32)
      loss_op = losses.BootstrappedSigmoidClassificationLoss(
          alpha, bootstrap_type='soft')
      loss = loss_op(prediction_tensor, target_tensor, weights=weights)
      loss = tf.reduce_sum(loss)
      return loss
    loss_output = self.execute(graph_fn, [])
    exp_loss = -math.log(.5)
    self.assertAllClose(loss_output, exp_loss)

  def testReturnsCorrectLossHardBootstrapping(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[-100, 100, 0],
                                        [100, -100, -100],
                                        [100, -100, -100],
                                        [-100, -100, 100]],
                                       [[-100, -100, 100],
                                        [-100, 100, -100],
                                        [100, 100, 100],
                                        [0, 0, -1]]], tf.float32)
      target_tensor = tf.constant([[[0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0],
                                    [0, 0, 1]],
                                   [[0, 0, 1],
                                    [0, 1, 0],
                                    [1, 1, 1],
                                    [1, 0, 0]]], tf.float32)
      weights = tf.constant([[[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [0, 0, 0]]], tf.float32)
      alpha = tf.constant(.5, tf.float32)
      loss_op = losses.BootstrappedSigmoidClassificationLoss(
          alpha, bootstrap_type='hard')
      loss = loss_op(prediction_tensor, target_tensor, weights=weights)
      loss = tf.reduce_sum(loss)
      return loss
    loss_output = self.execute(graph_fn, [])
    exp_loss = -math.log(.5)
    self.assertAllClose(loss_output, exp_loss)

  def testReturnsCorrectAnchorWiseLoss(self):
    def graph_fn():
      prediction_tensor = tf.constant([[[-100, 100, -100],
                                        [100, -100, -100],
                                        [100, 0, -100],
                                        [-100, -100, 100]],
                                       [[-100, 0, 100],
                                        [-100, 100, -100],
                                        [100, 100, 100],
                                        [0, 0, -1]]], tf.float32)
      target_tensor = tf.constant([[[0, 1, 0],
                                    [1, 0, 0],
                                    [1, 0, 0],
                                    [0, 0, 1]],
                                   [[0, 0, 1],
                                    [0, 1, 0],
                                    [1, 1, 1],
                                    [1, 0, 0]]], tf.float32)
      weights = tf.constant([[[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]],
                             [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1],
                              [0, 0, 0]]], tf.float32)
      alpha = tf.constant(.5, tf.float32)
      loss_op = losses.BootstrappedSigmoidClassificationLoss(
          alpha, bootstrap_type='hard')
      loss = loss_op(prediction_tensor, target_tensor, weights=weights)
      loss = tf.reduce_sum(loss, axis=2)
      return loss
    loss_output = self.execute(graph_fn, [])
    exp_loss = np.matrix([[0, 0, -math.log(.5), 0],
                          [-math.log(.5), 0, 0, 0]])
    self.assertAllClose(loss_output, exp_loss)


class HardExampleMinerTest(test_case.TestCase):

  def testHardMiningWithSingleLossType(self):
    def graph_fn():
      location_losses = tf.constant([[100, 90, 80, 0],
                                     [0, 1, 2, 3]], tf.float32)
      cls_losses = tf.constant([[0, 10, 50, 110],
                                [9, 6, 3, 0]], tf.float32)
      box_corners = tf.constant([[0.1, 0.1, 0.9, 0.9],
                                 [0.1, 0.1, 0.9, 0.9],
                                 [0.1, 0.1, 0.9, 0.9],
                                 [0.1, 0.1, 0.9, 0.9]], tf.float32)
      decoded_boxlist_list = []
      decoded_boxlist_list.append(box_list.BoxList(box_corners))
      decoded_boxlist_list.append(box_list.BoxList(box_corners))
      # Uses only location loss to select hard examples
      loss_op = losses.HardExampleMiner(num_hard_examples=1,
                                        iou_threshold=0.0,
                                        loss_type='loc',
                                        cls_loss_weight=1,
                                        loc_loss_weight=1)
      (loc_loss, cls_loss) = loss_op(location_losses, cls_losses,
                                     decoded_boxlist_list)
      return loc_loss, cls_loss
    loc_loss_output, cls_loss_output = self.execute(graph_fn, [])
    exp_loc_loss = 100 + 3
    exp_cls_loss = 0 + 0
    self.assertAllClose(loc_loss_output, exp_loc_loss)
    self.assertAllClose(cls_loss_output, exp_cls_loss)

  def testHardMiningWithBothLossType(self):
    def graph_fn():
      location_losses = tf.constant([[100, 90, 80, 0],
                                     [0, 1, 2, 3]], tf.float32)
      cls_losses = tf.constant([[0, 10, 50, 110],
                                [9, 6, 3, 0]], tf.float32)
      box_corners = tf.constant([[0.1, 0.1, 0.9, 0.9],
                                 [0.1, 0.1, 0.9, 0.9],
                                 [0.1, 0.1, 0.9, 0.9],
                                 [0.1, 0.1, 0.9, 0.9]], tf.float32)
      decoded_boxlist_list = []
      decoded_boxlist_list.append(box_list.BoxList(box_corners))
      decoded_boxlist_list.append(box_list.BoxList(box_corners))
      loss_op = losses.HardExampleMiner(num_hard_examples=1,
                                        iou_threshold=0.0,
                                        loss_type='both',
                                        cls_loss_weight=1,
                                        loc_loss_weight=1)
      (loc_loss, cls_loss) = loss_op(location_losses, cls_losses,
                                     decoded_boxlist_list)
      return loc_loss, cls_loss
    loc_loss_output, cls_loss_output = self.execute(graph_fn, [])
    exp_loc_loss = 80 + 0
    exp_cls_loss = 50 + 9
    self.assertAllClose(loc_loss_output, exp_loc_loss)
    self.assertAllClose(cls_loss_output, exp_cls_loss)

  def testHardMiningNMS(self):
    def graph_fn():
      location_losses = tf.constant([[100, 90, 80, 0],
                                     [0, 1, 2, 3]], tf.float32)
      cls_losses = tf.constant([[0, 10, 50, 110],
                                [9, 6, 3, 0]], tf.float32)
      box_corners = tf.constant([[0.1, 0.1, 0.9, 0.9],
                                 [0.9, 0.9, 0.99, 0.99],
                                 [0.1, 0.1, 0.9, 0.9],
                                 [0.1, 0.1, 0.9, 0.9]], tf.float32)
      decoded_boxlist_list = []
      decoded_boxlist_list.append(box_list.BoxList(box_corners))
      decoded_boxlist_list.append(box_list.BoxList(box_corners))
      loss_op = losses.HardExampleMiner(num_hard_examples=2,
                                        iou_threshold=0.5,
                                        loss_type='cls',
                                        cls_loss_weight=1,
                                        loc_loss_weight=1)
      (loc_loss, cls_loss) = loss_op(location_losses, cls_losses,
                                     decoded_boxlist_list)
      return loc_loss, cls_loss
    loc_loss_output, cls_loss_output = self.execute(graph_fn, [])
    exp_loc_loss = 0 + 90 + 0 + 1
    exp_cls_loss = 110 + 10 + 9 + 6

    self.assertAllClose(loc_loss_output, exp_loc_loss)
    self.assertAllClose(cls_loss_output, exp_cls_loss)

  def testEnforceNegativesPerPositiveRatio(self):
    location_losses = np.array([[100, 90, 80, 0, 1, 2,
                                 3, 10, 20, 100, 20, 3]], np.float32)
    cls_losses = np.array([[0, 0, 100, 0, 90, 70,
                            0, 60, 0, 17, 13, 0]], np.float32)
    box_corners = np.array([[0.0, 0.0, 0.2, 0.1],
                            [0.0, 0.0, 0.2, 0.1],
                            [0.0, 0.0, 0.2, 0.1],
                            [0.0, 0.0, 0.2, 0.1],
                            [0.0, 0.0, 0.5, 0.1],
                            [0.0, 0.0, 0.6, 0.1],
                            [0.0, 0.0, 0.2, 0.1],
                            [0.0, 0.0, 0.8, 0.1],
                            [0.0, 0.0, 0.2, 0.1],
                            [0.0, 0.0, 1.0, 0.1],
                            [0.0, 0.0, 1.1, 0.1],
                            [0.0, 0.0, 0.2, 0.1]], np.float32)
    match_results = np.array([2, -1, 0, -1, -1, 1, -1, -1, -1, -1, -1, 3],
                             np.int32)

    max_negatives_per_positive_list = [0.0, 0.5, 1.0, 1.5, 10]
    exp_loc_loss_list = [80 + 2,
                         80 + 1 + 2,
                         80 + 1 + 2 + 10,
                         80 + 1 + 2 + 10 + 100,
                         80 + 1 + 2 + 10 + 100 + 20]
    exp_cls_loss_list = [100 + 70,
                         100 + 90 + 70,
                         100 + 90 + 70 + 60,
                         100 + 90 + 70 + 60 + 17,
                         100 + 90 + 70 + 60 + 17 + 13]

    # pylint: disable=cell-var-from-loop
    for max_negatives_per_positive, exp_loc_loss, exp_cls_loss in zip(
        max_negatives_per_positive_list, exp_loc_loss_list, exp_cls_loss_list):
      def graph_fn():
        loss_op = losses.HardExampleMiner(
            num_hard_examples=None, iou_threshold=0.9999, loss_type='cls',
            cls_loss_weight=1, loc_loss_weight=1,
            max_negatives_per_positive=max_negatives_per_positive)
        match_list = [matcher.Match(tf.constant(match_results))]
        decoded_boxlist_list = [box_list.BoxList(tf.constant(box_corners))]
        (loc_loss, cls_loss) = loss_op(tf.constant(location_losses),
                                       tf.constant(cls_losses),
                                       decoded_boxlist_list, match_list)
        return loc_loss, cls_loss
      loc_loss_output, cls_loss_output = self.execute_cpu(graph_fn, [])
      self.assertAllClose(loc_loss_output, exp_loc_loss)
      self.assertAllClose(cls_loss_output, exp_cls_loss)
    # pylint: enable=cell-var-from-loop

  def testEnforceNegativesPerPositiveRatioWithMinNegativesPerImage(self):
    location_losses = np.array([[100, 90, 80, 0, 1, 2,
                                 3, 10, 20, 100, 20, 3]], np.float32)
    cls_losses = np.array([[0, 0, 100, 0, 90, 70,
                            0, 60, 0, 17, 13, 0]], np.float32)
    box_corners = np.array([[0.0, 0.0, 0.2, 0.1],
                            [0.0, 0.0, 0.2, 0.1],
                            [0.0, 0.0, 0.2, 0.1],
                            [0.0, 0.0, 0.2, 0.1],
                            [0.0, 0.0, 0.5, 0.1],
                            [0.0, 0.0, 0.6, 0.1],
                            [0.0, 0.0, 0.2, 0.1],
                            [0.0, 0.0, 0.8, 0.1],
                            [0.0, 0.0, 0.2, 0.1],
                            [0.0, 0.0, 1.0, 0.1],
                            [0.0, 0.0, 1.1, 0.1],
                            [0.0, 0.0, 0.2, 0.1]], np.float32)
    match_results = np.array([-1] * 12, np.int32)

    min_negatives_per_image_list = [0, 1, 2, 4, 5, 6]
    exp_loc_loss_list = [0,
                         80,
                         80 + 1,
                         80 + 1 + 2 + 10,
                         80 + 1 + 2 + 10 + 100,
                         80 + 1 + 2 + 10 + 100 + 20]
    exp_cls_loss_list = [0,
                         100,
                         100 + 90,
                         100 + 90 + 70 + 60,
                         100 + 90 + 70 + 60 + 17,
                         100 + 90 + 70 + 60 + 17 + 13]

    # pylint: disable=cell-var-from-loop
    for min_negatives_per_image, exp_loc_loss, exp_cls_loss in zip(
        min_negatives_per_image_list, exp_loc_loss_list, exp_cls_loss_list):
      def graph_fn():
        loss_op = losses.HardExampleMiner(
            num_hard_examples=None, iou_threshold=0.9999, loss_type='cls',
            cls_loss_weight=1, loc_loss_weight=1,
            max_negatives_per_positive=3,
            min_negatives_per_image=min_negatives_per_image)
        match_list = [matcher.Match(tf.constant(match_results))]
        decoded_boxlist_list = [box_list.BoxList(tf.constant(box_corners))]
        (loc_loss, cls_loss) = loss_op(location_losses, cls_losses,
                                       decoded_boxlist_list, match_list)
        return loc_loss, cls_loss
      loc_loss_output, cls_loss_output = self.execute_cpu(graph_fn, [])
      self.assertAllClose(loc_loss_output, exp_loc_loss)
      self.assertAllClose(cls_loss_output, exp_cls_loss)
    # pylint: enable=cell-var-from-loop


LOG_2 = np.log(2)
LOG_3 = np.log(3)


class PenaltyReducedLogisticFocalLossTest(test_case.TestCase):
  """Testing loss function from Equation (1) in [1].

  [1]: https://arxiv.org/abs/1904.07850
  """

  def setUp(self):
    super(PenaltyReducedLogisticFocalLossTest, self).setUp()
    self._prediction = np.array([
        # First batch
        [[1 / 2, 1 / 4, 3 / 4],
         [3 / 4, 1 / 3, 1 / 3]],
        # Second Batch
        [[0.0, 1.0, 1 / 2],
         [3 / 4, 2 / 3, 1 / 3]]], np.float32)
    self._prediction = np.log(self._prediction/(1 - self._prediction))

    self._target = np.array([
        # First batch
        [[1.0, 0.91, 1.0],
         [0.36, 0.84, 1.0]],
        # Second Batch
        [[0.01, 1.0, 0.75],
         [0.96, 1.0, 1.0]]], np.float32)

  def test_returns_correct_loss(self):
    def graph_fn(prediction, target):
      weights = tf.constant([
          [[1.0], [1.0]],
          [[1.0], [1.0]],
      ])
      loss = losses.PenaltyReducedLogisticFocalLoss(alpha=2.0, beta=0.5)
      computed_value = loss._compute_loss(prediction, target,
                                          weights)
      return computed_value
    computed_value = self.execute(graph_fn, [self._prediction, self._target])
    expected_value = np.array([
        # First batch
        [[1 / 4 * LOG_2,
          0.3 * 0.0625 * (2 * LOG_2 - LOG_3),
          1 / 16 * (2 * LOG_2 - LOG_3)],
         [0.8 * 9 / 16 * 2 * LOG_2,
          0.4 * 1 / 9 * (LOG_3 - LOG_2),
          4 / 9 * LOG_3]],
        # Second Batch
        [[0.0,
          0.0,
          1 / 2 * 1 / 4 * LOG_2],
         [0.2 * 9 / 16 * 2 * LOG_2,
          1 / 9 * (LOG_3 - LOG_2),
          4 / 9 * LOG_3]]])
    self.assertAllClose(computed_value, expected_value, rtol=1e-3, atol=1e-3)

  def test_returns_correct_loss_weighted(self):
    def graph_fn(prediction, target):
      weights = tf.constant([
          [[1.0, 0.0, 1.0], [0.0, 0.0, 1.0]],
          [[1.0, 1.0, 1.0], [0.0, 0.0, 0.0]],
      ])

      loss = losses.PenaltyReducedLogisticFocalLoss(alpha=2.0, beta=0.5)

      computed_value = loss._compute_loss(prediction, target,
                                          weights)
      return computed_value
    computed_value = self.execute(graph_fn, [self._prediction, self._target])
    expected_value = np.array([
        # First batch
        [[1 / 4 * LOG_2,
          0.0,
          1 / 16 * (2 * LOG_2 - LOG_3)],
         [0.0,
          0.0,
          4 / 9 * LOG_3]],
        # Second Batch
        [[0.0,
          0.0,
          1 / 2 * 1 / 4 * LOG_2],
         [0.0,
          0.0,
          0.0]]])

    self.assertAllClose(computed_value, expected_value, rtol=1e-3, atol=1e-3)


class L1LocalizationLossTest(test_case.TestCase):

  def test_returns_correct_loss(self):
    def graph_fn():
      loss = losses.L1LocalizationLoss()
      pred = [[0.1, 0.2], [0.7, 0.5]]
      target = [[0.9, 1.0], [0.1, 0.4]]

      weights = [[1.0, 0.0], [1.0, 1.0]]
      return loss._compute_loss(pred, target, weights)
    computed_value = self.execute(graph_fn, [])
    self.assertAllClose(computed_value, [[0.8, 0.0], [0.6, 0.1]], rtol=1e-6)


class WeightedDiceClassificationLoss(test_case.TestCase):

  def test_compute_weights_1(self):
    def graph_fn():
      loss = losses.WeightedDiceClassificationLoss(squared_normalization=False)
      pred = np.zeros((2, 3, 4), dtype=np.float32)
      target = np.zeros((2, 3, 4), dtype=np.float32)

      pred[0, 1, 0] = _logit(0.9)
      pred[0, 2, 0] = _logit(0.1)
      pred[0, 2, 2] = _logit(0.5)
      pred[0, 1, 3] = _logit(0.1)

      pred[1, 2, 3] = _logit(0.2)
      pred[1, 1, 1] = _logit(0.3)
      pred[1, 0, 2] = _logit(0.1)

      target[0, 1, 0] = 1.0
      target[0, 2, 2] = 1.0
      target[0, 1, 3] = 1.0

      target[1, 2, 3] = 1.0
      target[1, 1, 1] = 0.0
      target[1, 0, 2] = 0.0

      weights = np.ones_like(target)
      return loss._compute_loss(pred, target, weights)

    dice_coeff = np.zeros((2, 4))
    dice_coeff[0, 0] = 2 * 0.9 / 2.5
    dice_coeff[0, 2] = 2 * 0.5 / 2.5
    dice_coeff[0, 3] = 2 * 0.1 / 2.1
    dice_coeff[1, 3] = 2 * 0.2 / 2.2

    computed_value = self.execute(graph_fn, [])
    self.assertAllClose(computed_value, 1 - dice_coeff, rtol=1e-6)

  def test_compute_weights_set(self):

    def graph_fn():
      loss = losses.WeightedDiceClassificationLoss(squared_normalization=False)
      pred = np.zeros((2, 3, 4), dtype=np.float32)
      target = np.zeros((2, 3, 4), dtype=np.float32)

      pred[0, 1, 0] = _logit(0.9)
      pred[0, 2, 0] = _logit(0.1)
      pred[0, 2, 2] = _logit(0.5)
      pred[0, 1, 3] = _logit(0.1)

      pred[1, 2, 3] = _logit(0.2)
      pred[1, 1, 1] = _logit(0.3)
      pred[1, 0, 2] = _logit(0.1)

      target[0, 1, 0] = 1.0
      target[0, 2, 2] = 1.0
      target[0, 1, 3] = 1.0

      target[1, 2, 3] = 1.0
      target[1, 1, 1] = 0.0
      target[1, 0, 2] = 0.0

      weights = np.ones_like(target)
      weights[:, :, 0] = 0.0
      return loss._compute_loss(pred, target, weights)

    dice_coeff = np.zeros((2, 4))
    dice_coeff[0, 2] = 2 * 0.5 / 2.5
    dice_coeff[0, 3] = 2 * 0.1 / 2.1
    dice_coeff[1, 3] = 2 * 0.2 / 2.2

    computed_value = self.execute(graph_fn, [])
    self.assertAllClose(computed_value, 1 - dice_coeff, rtol=1e-6)

  def test_class_indices(self):
    def graph_fn():
      loss = losses.WeightedDiceClassificationLoss(squared_normalization=False)
      pred = np.zeros((2, 3, 4), dtype=np.float32)
      target = np.zeros((2, 3, 4), dtype=np.float32)

      pred[0, 1, 0] = _logit(0.9)
      pred[0, 2, 0] = _logit(0.1)
      pred[0, 2, 2] = _logit(0.5)
      pred[0, 1, 3] = _logit(0.1)

      pred[1, 2, 3] = _logit(0.2)
      pred[1, 1, 1] = _logit(0.3)
      pred[1, 0, 2] = _logit(0.1)

      target[0, 1, 0] = 1.0
      target[0, 2, 2] = 1.0
      target[0, 1, 3] = 1.0

      target[1, 2, 3] = 1.0
      target[1, 1, 1] = 0.0
      target[1, 0, 2] = 0.0

      weights = np.ones_like(target)
      return loss._compute_loss(pred, target, weights, class_indices=[0])

    dice_coeff = np.zeros((2, 4))
    dice_coeff[0, 0] = 2 * 0.9 / 2.5

    computed_value = self.execute(graph_fn, [])
    self.assertAllClose(computed_value, 1 - dice_coeff, rtol=1e-6)


if __name__ == '__main__':
  tf.test.main()



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Interface for data parsers.

Data parser parses input data and returns a dictionary of numpy arrays
keyed by the entries in standard_fields.py. Since the parser parses records
to numpy arrays (materialized tensors) directly, it is used to read data for
evaluation/visualization; to parse the data during training, DataDecoder should
be used.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from abc import ABCMeta
from abc import abstractmethod
import six


class DataToNumpyParser(six.with_metaclass(ABCMeta, object)):
  """Abstract interface for data parser that produces numpy arrays."""

  @abstractmethod
  def parse(self, input_data):
    """Parses input and returns a numpy array or a dictionary of numpy arrays.

    Args:
      input_data: an input data

    Returns:
      A numpy array or a dictionary of numpy arrays or None, if input
      cannot be parsed.
    """
    pass



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tensorflow Example proto parser for data loading.

A parser to decode data containing serialized tensorflow.Example
protos into materialized tensors (numpy arrays).
"""

import numpy as np

from object_detection.core import data_parser
from object_detection.core import standard_fields as fields


class FloatParser(data_parser.DataToNumpyParser):
  """Tensorflow Example float parser."""

  def __init__(self, field_name):
    self.field_name = field_name

  def parse(self, tf_example):
    return np.array(
        tf_example.features.feature[self.field_name].float_list.value,
        dtype=np.float).transpose() if tf_example.features.feature[
            self.field_name].HasField("float_list") else None


class StringParser(data_parser.DataToNumpyParser):
  """Tensorflow Example string parser."""

  def __init__(self, field_name):
    self.field_name = field_name

  def parse(self, tf_example):
    return b"".join(tf_example.features.feature[
        self.field_name].bytes_list.value) if tf_example.features.feature[
            self.field_name].HasField("bytes_list") else None


class Int64Parser(data_parser.DataToNumpyParser):
  """Tensorflow Example int64 parser."""

  def __init__(self, field_name):
    self.field_name = field_name

  def parse(self, tf_example):
    return np.array(
        tf_example.features.feature[self.field_name].int64_list.value,
        dtype=np.int64).transpose() if tf_example.features.feature[
            self.field_name].HasField("int64_list") else None


class BoundingBoxParser(data_parser.DataToNumpyParser):
  """Tensorflow Example bounding box parser."""

  def __init__(self, xmin_field_name, ymin_field_name, xmax_field_name,
               ymax_field_name):
    self.field_names = [
        ymin_field_name, xmin_field_name, ymax_field_name, xmax_field_name
    ]

  def parse(self, tf_example):
    result = []
    parsed = True
    for field_name in self.field_names:
      result.append(tf_example.features.feature[field_name].float_list.value)
      parsed &= (
          tf_example.features.feature[field_name].HasField("float_list"))

    return np.array(result).transpose() if parsed else None


class TfExampleDetectionAndGTParser(data_parser.DataToNumpyParser):
  """Tensorflow Example proto parser."""

  def __init__(self):
    self.items_to_handlers = {
        fields.DetectionResultFields.key:
            StringParser(fields.TfExampleFields.source_id),
        # Object ground truth boxes and classes.
        fields.InputDataFields.groundtruth_boxes: (BoundingBoxParser(
            fields.TfExampleFields.object_bbox_xmin,
            fields.TfExampleFields.object_bbox_ymin,
            fields.TfExampleFields.object_bbox_xmax,
            fields.TfExampleFields.object_bbox_ymax)),
        fields.InputDataFields.groundtruth_classes: (
            Int64Parser(fields.TfExampleFields.object_class_label)),
        # Object detections.
        fields.DetectionResultFields.detection_boxes: (BoundingBoxParser(
            fields.TfExampleFields.detection_bbox_xmin,
            fields.TfExampleFields.detection_bbox_ymin,
            fields.TfExampleFields.detection_bbox_xmax,
            fields.TfExampleFields.detection_bbox_ymax)),
        fields.DetectionResultFields.detection_classes: (
            Int64Parser(fields.TfExampleFields.detection_class_label)),
        fields.DetectionResultFields.detection_scores: (
            FloatParser(fields.TfExampleFields.detection_score)),
    }

    self.optional_items_to_handlers = {
        fields.InputDataFields.groundtruth_difficult:
            Int64Parser(fields.TfExampleFields.object_difficult),
        fields.InputDataFields.groundtruth_group_of:
            Int64Parser(fields.TfExampleFields.object_group_of),
        fields.InputDataFields.groundtruth_image_classes:
            Int64Parser(fields.TfExampleFields.image_class_label),
    }

  def parse(self, tf_example):
    """Parses tensorflow example and returns a tensor dictionary.

    Args:
      tf_example: a tf.Example object.

    Returns:
      A dictionary of the following numpy arrays:
      fields.DetectionResultFields.source_id - string containing original image
      id.
      fields.InputDataFields.groundtruth_boxes - a numpy array containing
      groundtruth boxes.
      fields.InputDataFields.groundtruth_classes - a numpy array containing
      groundtruth classes.
      fields.InputDataFields.groundtruth_group_of - a numpy array containing
      groundtruth group of flag (optional, None if not specified).
      fields.InputDataFields.groundtruth_difficult - a numpy array containing
      groundtruth difficult flag (optional, None if not specified).
      fields.InputDataFields.groundtruth_image_classes - a numpy array
      containing groundtruth image-level labels.
      fields.DetectionResultFields.detection_boxes - a numpy array containing
      detection boxes.
      fields.DetectionResultFields.detection_classes - a numpy array containing
      detection class labels.
      fields.DetectionResultFields.detection_scores - a numpy array containing
      detection scores.
      Returns None if tf.Example was not parsed or non-optional fields were not
      found.
    """
    results_dict = {}
    parsed = True
    for key, parser in self.items_to_handlers.items():
      results_dict[key] = parser.parse(tf_example)
      parsed &= (results_dict[key] is not None)

    for key, parser in self.optional_items_to_handlers.items():
      results_dict[key] = parser.parse(tf_example)

    return results_dict if parsed else None



# Copyright (c) Facebook, Inc. and its affiliates.
import json
import logging
import os

from detectron2.data import DatasetCatalog, MetadataCatalog
from detectron2.data.datasets.builtin_meta import CITYSCAPES_CATEGORIES
from detectron2.utils.file_io import PathManager

"""
This file contains functions to register the Cityscapes panoptic dataset to the DatasetCatalog.
"""


logger = logging.getLogger(__name__)


def get_cityscapes_panoptic_files(image_dir, gt_dir, json_info):
    files = []
    # scan through the directory
    cities = PathManager.ls(image_dir)
    logger.info(f"{len(cities)} cities found in '{image_dir}'.")
    image_dict = {}
    for city in cities:
        city_img_dir = os.path.join(image_dir, city)
        for basename in PathManager.ls(city_img_dir):
            image_file = os.path.join(city_img_dir, basename)

            suffix = "_leftImg8bit.png"
            assert basename.endswith(suffix), basename
            basename = os.path.basename(basename)[: -len(suffix)]

            image_dict[basename] = image_file

    for ann in json_info["annotations"]:
        image_file = image_dict.get(ann["image_id"], None)
        assert image_file is not None, "No image {} found for annotation {}".format(
            ann["image_id"], ann["file_name"]
        )
        label_file = os.path.join(gt_dir, ann["file_name"])
        segments_info = ann["segments_info"]

        files.append((image_file, label_file, segments_info))

    assert len(files), "No images found in {}".format(image_dir)
    assert PathManager.isfile(files[0][0]), files[0][0]
    assert PathManager.isfile(files[0][1]), files[0][1]
    return files


def load_cityscapes_panoptic(image_dir, gt_dir, gt_json, meta):
    """
    Args:
        image_dir (str): path to the raw dataset. e.g., "~/cityscapes/leftImg8bit/train".
        gt_dir (str): path to the raw annotations. e.g.,
            "~/cityscapes/gtFine/cityscapes_panoptic_train".
        gt_json (str): path to the json file. e.g.,
            "~/cityscapes/gtFine/cityscapes_panoptic_train.json".
        meta (dict): dictionary containing "thing_dataset_id_to_contiguous_id"
            and "stuff_dataset_id_to_contiguous_id" to map category ids to
            contiguous ids for training.

    Returns:
        list[dict]: a list of dicts in Detectron2 standard format. (See
        `Using Custom Datasets </tutorials/datasets.html>`_ )
    """

    def _convert_category_id(segment_info, meta):
        if segment_info["category_id"] in meta["thing_dataset_id_to_contiguous_id"]:
            segment_info["category_id"] = meta["thing_dataset_id_to_contiguous_id"][
                segment_info["category_id"]
            ]
        else:
            segment_info["category_id"] = meta["stuff_dataset_id_to_contiguous_id"][
                segment_info["category_id"]
            ]
        return segment_info

    assert os.path.exists(
        gt_json
    ), "Please run `python cityscapesscripts/preparation/createPanopticImgs.py` to generate label files."  # noqa
    with open(gt_json) as f:
        json_info = json.load(f)
    files = get_cityscapes_panoptic_files(image_dir, gt_dir, json_info)
    ret = []
    for image_file, label_file, segments_info in files:
        sem_label_file = (
            image_file.replace("leftImg8bit", "gtFine").split(".")[0] + "_labelTrainIds.png"
        )
        segments_info = [_convert_category_id(x, meta) for x in segments_info]
        ret.append(
            {
                "file_name": image_file,
                "image_id": "_".join(
                    os.path.splitext(os.path.basename(image_file))[0].split("_")[:3]
                ),
                "sem_seg_file_name": sem_label_file,
                "pan_seg_file_name": label_file,
                "segments_info": segments_info,
            }
        )
    assert len(ret), f"No images found in {image_dir}!"
    assert PathManager.isfile(
        ret[0]["sem_seg_file_name"]
    ), "Please generate labelTrainIds.png with cityscapesscripts/preparation/createTrainIdLabelImgs.py"  # noqa
    assert PathManager.isfile(
        ret[0]["pan_seg_file_name"]
    ), "Please generate panoptic annotation with python cityscapesscripts/preparation/createPanopticImgs.py"  # noqa
    return ret


_RAW_CITYSCAPES_PANOPTIC_SPLITS = {
    "cityscapes_fine_panoptic_train": (
        "cityscapes/leftImg8bit/train",
        "cityscapes/gtFine/cityscapes_panoptic_train",
        "cityscapes/gtFine/cityscapes_panoptic_train.json",
    ),
    "cityscapes_fine_panoptic_val": (
        "cityscapes/leftImg8bit/val",
        "cityscapes/gtFine/cityscapes_panoptic_val",
        "cityscapes/gtFine/cityscapes_panoptic_val.json",
    ),
    # "cityscapes_fine_panoptic_test": not supported yet
}


def register_all_cityscapes_panoptic(root):
    meta = {}
    # The following metadata maps contiguous id from [0, #thing categories +
    # #stuff categories) to their names and colors. We have to replica of the
    # same name and color under "thing_*" and "stuff_*" because the current
    # visualization function in D2 handles thing and class classes differently
    # due to some heuristic used in Panoptic FPN. We keep the same naming to
    # enable reusing existing visualization functions.
    thing_classes = [k["name"] for k in CITYSCAPES_CATEGORIES]
    thing_colors = [k["color"] for k in CITYSCAPES_CATEGORIES]
    stuff_classes = [k["name"] for k in CITYSCAPES_CATEGORIES]
    stuff_colors = [k["color"] for k in CITYSCAPES_CATEGORIES]

    meta["thing_classes"] = thing_classes
    meta["thing_colors"] = thing_colors
    meta["stuff_classes"] = stuff_classes
    meta["stuff_colors"] = stuff_colors

    # There are three types of ids in cityscapes panoptic segmentation:
    # (1) category id: like semantic segmentation, it is the class id for each
    #   pixel. Since there are some classes not used in evaluation, the category
    #   id is not always contiguous and thus we have two set of category ids:
    #       - original category id: category id in the original dataset, mainly
    #           used for evaluation.
    #       - contiguous category id: [0, #classes), in order to train the classifier
    # (2) instance id: this id is used to differentiate different instances from
    #   the same category. For "stuff" classes, the instance id is always 0; for
    #   "thing" classes, the instance id starts from 1 and 0 is reserved for
    #   ignored instances (e.g. crowd annotation).
    # (3) panoptic id: this is the compact id that encode both category and
    #   instance id by: category_id * 1000 + instance_id.
    thing_dataset_id_to_contiguous_id = {}
    stuff_dataset_id_to_contiguous_id = {}

    for k in CITYSCAPES_CATEGORIES:
        if k["isthing"] == 1:
            thing_dataset_id_to_contiguous_id[k["id"]] = k["trainId"]
        else:
            stuff_dataset_id_to_contiguous_id[k["id"]] = k["trainId"]

    meta["thing_dataset_id_to_contiguous_id"] = thing_dataset_id_to_contiguous_id
    meta["stuff_dataset_id_to_contiguous_id"] = stuff_dataset_id_to_contiguous_id

    for key, (image_dir, gt_dir, gt_json) in _RAW_CITYSCAPES_PANOPTIC_SPLITS.items():
        image_dir = os.path.join(root, image_dir)
        gt_dir = os.path.join(root, gt_dir)
        gt_json = os.path.join(root, gt_json)

        DatasetCatalog.register(
            key, lambda x=image_dir, y=gt_dir, z=gt_json: load_cityscapes_panoptic(x, y, z, meta)
        )
        MetadataCatalog.get(key).set(
            panoptic_root=gt_dir,
            image_root=image_dir,
            panoptic_json=gt_json,
            gt_dir=gt_dir.replace("cityscapes_panoptic_", ""),
            evaluator_type="cityscapes_panoptic_seg",
            ignore_label=255,
            label_divisor=1000,
            **meta,
        )
